# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The versioning module contains the version manager for the Scade One formats.
"""

import re
import json
from pathlib import Path
from ansys.scadeone.core.common.exception import ScadeOneException

VersionFile = Path(__file__).parent / "versions.json"


class VersionManager:
    """Class managing the versions for the Scade One tools formats.

    The version manager is used through the singleton `FormatVersions` instance.
    """

    VersionRE = re.compile(r"(?P<M>\d+)\.(?P<m>\d+)")

    def __init__(self) -> None:
        self._format_versions = {}
        self._formats = None

    @property
    def formats(self) -> set:
        """Get the formats as a set of strings."""
        if self._formats is None:
            self._formats = set(self._format_versions.keys())
        return self._formats

    def get_versions(self) -> str:
        """Get the versions as a document string.

        Returns
        -------
        str
            String containing the versions
        """
        buffer = "The versions for the supported format/code are:\n\n"
        for k in sorted(self.formats):
            buffer += f"- {self.description(k)}: {self.version(k)}\n"
        return buffer

    def load_versions(self, version_file: Path):
        """Load the versions from the json file file.

        Parameters
        ----------
        version_file : Path
            JSON file containing the versions

        Raises
        ------
        ScadeOneException
            If the file is not found or if the file is not valid.
        """
        try:
            data = version_file.read_text()
            self._format_versions = json.loads(data)
        except Exception as e:
            raise ScadeOneException(f"Unable to load versions: {e}")
        if not isinstance(self._format_versions, dict):
            raise ScadeOneException(f"Invalid versions file: {version_file}")
        for k, v in self._format_versions.items():
            if not (isinstance(v, dict) and "version" in v and "description" in v):
                raise ScadeOneException(f"{version_file}: Invalid version for {k}")

    def version(self, format: str):
        """Get the version for the format.

        Parameters
        ----------
        format : str
            Format name

        Returns
        -------
        str
            Version
        """
        if format not in self.formats:
            raise ScadeOneException(f"Unknown format: {format}")
        return self._format_versions[format]["version"]

    def description(self, format: str):
        """Get the description for the format.

        Parameters
        ----------
        format : str
            Format name

        Returns
        -------
        str
            Description
        """
        if format not in self.formats:
            raise ScadeOneException(f"Unknown format: {format}")
        return self._format_versions[format]["description"]

    def check(self, format: str, version: str):
        """Check if the version is supported for the format.

        Parameters
        ----------
        format : str
            Format name
        version : str
            Version to be checked

        Raises
        ------
        ScadeOneException
            If the version is not supported for the format.
        """
        if format not in self.formats:
            raise ScadeOneException(f"Unknown format: {format}")
        version_ref = self._format_versions[format]["version"]
        # check expected version is valid
        m = re.match(VersionManager.VersionRE, version_ref)
        if not m:
            raise ScadeOneException(f"Invalid {format} version, check installation: {version_ref}")
        expected = m.groupdict()
        # check provided version is valid
        m = re.match(VersionManager.VersionRE, version)
        if not m:
            raise ScadeOneException(f"Invalid {format} version: {version}, expecting {version_ref}")
        provided = m.groupdict()
        # check version
        if provided["M"] != expected["M"] or int(provided["m"]) > int(expected["m"]):
            raise ScadeOneException(
                f"Unsupported {format} version: {version}, expecting {version_ref}"
            )


FormatVersions = VersionManager()
FormatVersions.load_versions(VersionFile)


def gen_swan_version(is_harness=False):
    """Generate the version for a Swan file.

    Parameters
    ----------
    is_harness : bool
        True if the harness version is needed

    Returns
    -------
    str
        Version
    """
    version = "-- version"
    version += f" swan: {FormatVersions.version('swan')}"
    version += f" graph: {FormatVersions.version('graph')}"
    if is_harness:
        version += f" swant: {FormatVersions.version('swant')}"
    return version



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-FileCopyrightText: 2024 ANSYS, Inc.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING

from .diagram_creator import DiagramFactory
from .module_creator import DeclarationFactory, ModuleFactory
from .operator_creator import OperatorFactory
from .project_creator import ProjectFactory

if TYPE_CHECKING:
    pass


class ScadeOneFactory:
    """Factory for creating ScadeOne objects."""

    _instance = None

    def __init__(self) -> None:
        self.project = ProjectFactory()
        self.module = ModuleFactory()
        self.decl = DeclarationFactory()
        self.operator = OperatorFactory()
        self.diagram = DiagramFactory()

    def __new__(cls, *args, **kwargs) -> "ScadeOneFactory":
        if not cls._instance:
            cls._instance = super(ScadeOneFactory, cls).__new__(cls)
        return cls._instance



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains the classes for:

- var section
- let section
- emit section
- assume section
- guaranteed section
"""

from typing import List, Optional, Union

import ansys.scadeone.core.swan.common as common
import ansys.scadeone.core.swan.scopes as scopes

from .variable import VarDecl


class LetSection(scopes.ScopeSection):  # numpydoc ignore=PR01
    """Implements:

    **let** {{*equation* ;}} section.
    """

    def __init__(self, equations: List[common.Equation]) -> None:
        super().__init__()
        self._equations = equations
        common.SwanItem.set_owner(self, equations)

    @property
    def equations(self) -> List[common.Equation]:
        """List of equation in **let**."""
        return self._equations


class VarSection(scopes.ScopeSection):  # numpydoc ignore=PR01
    """Implements:

    **var** {{*var_decl* ;}} section."""

    def __init__(self, var_decls: List[VarDecl]) -> None:
        super().__init__()
        self._var_decls = var_decls
        common.SwanItem.set_owner(self, var_decls)

    @property
    def var_decls(self) -> List[VarDecl]:
        """Declared variables."""
        return self._var_decls


class EmissionBody(common.SwanItem):  # numpydoc ignore=PR01
    """Implements an emission:

    | *emission_body* ::= *flow_names* [[ **if** *expr* ]]
    | *flow_names* ::= NAME {{ , NAME }}
    """

    def __init__(
        self,
        flows: List[common.Identifier],
        condition: Optional[common.Expression] = None,
        luid: Optional[common.Luid] = None,
    ) -> None:
        super().__init__()
        self._flows = flows
        self._condition = condition
        self._luid = luid

    @property
    def flows(self) -> List[common.Identifier]:
        """Emitted flows."""
        return self._flows

    @property
    def condition(self) -> Union[common.Expression, None]:
        """Emission condition if exists, else None."""
        return self._condition

    @property
    def luid(self) -> Union[common.Luid, None]:
        """Emission identifier if exists, else None."""
        return self._luid


class EmitSection(scopes.ScopeSection):  # numpydoc ignore=PR01
    """Implements an Emit section:

    **emit** {{*emission_body* ;}}"""

    def __init__(self, emissions: List[EmissionBody]) -> None:
        super().__init__()
        self._emissions = emissions
        common.SwanItem.set_owner(self, emissions)

    @property
    def emissions(self) -> List[EmissionBody]:
        """List of emissions."""
        return self._emissions


class FormalProperty(common.SwanItem):  # numpydoc ignore=PR01
    """Assume or Guarantee expression."""

    def __init__(self, luid: common.Luid, expr: common.Expression) -> None:
        super().__init__()
        self._luid = luid
        self._expr = expr

    @property
    def luid(self) -> common.Luid:
        """Property identifier."""
        return self._luid

    @property
    def expr(self) -> common.Expression:
        """Property expression."""
        return self._expr


class AssertSection(scopes.ScopeSection):  # numpydoc ignore=PR01
    """Implements Assert section:

    **assert** {{LUID: *expr* ;}}"""

    def __init__(self, assertions: List[FormalProperty]) -> None:
        super().__init__()
        self._assertions = assertions
        common.SwanItem.set_owner(self, assertions)

    @property
    def assertions(self) -> List[FormalProperty]:
        """Hypotheses of Assert."""
        return self._assertions


class AssumeSection(scopes.ScopeSection):  # numpydoc ignore=PR01
    """Implements Assume section:

    **assume** {{LUID: *expr* ;}}"""

    def __init__(self, hypotheses: List[FormalProperty]) -> None:
        super().__init__()
        self._hypotheses = hypotheses
        common.SwanItem.set_owner(self, hypotheses)

    @property
    def hypotheses(self) -> List[FormalProperty]:
        """Hypotheses of Assume."""
        return self._hypotheses


class GuaranteeSection(scopes.ScopeSection):  # numpydoc ignore=PR01
    """Implements Guarantee section:

    **guarantee** {{LUID: *expr* ;}}"""

    def __init__(self, guarantees: List[FormalProperty]) -> None:
        super().__init__()
        self._guarantees = guarantees
        common.SwanItem.set_owner(self, guarantees)

    @property
    def guarantees(self) -> List[FormalProperty]:
        """Guarantees of Guarantee."""
        return self._guarantees


class ProtectedSection(scopes.ScopeSection, common.ProtectedItem):  # numpydoc ignore=PR01
    """Protected section, meaning a syntactically incorrect section construct."""

    def __init__(self, data: str) -> None:
        scopes.ScopeSection.__init__(self)
        common.ProtectedItem.__init__(self, data)


# Diagram section is in diagram.py



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import ABC, abstractmethod
from enum import Enum, auto
import subprocess
from pathlib import Path
from typing import Optional, Union, TypedDict
import uuid

from ansys.scadeone.core.common.storage import JobFile, JobStorage

from ansys.scadeone.core.common.versioning import FormatVersions
from ansys.scadeone.core.interfaces import IProject


class _PropertiesData(TypedDict, total=False):
    """Job properties data structure."""

    # Common fields
    RootDeclarations: list[str]
    Name: str
    CustomArguments: str

    # Code generation fields
    Expansion: str
    ExpansionExp: str
    ExpansionNoExp: str
    ShortCircuitOperators: str
    NameLength: str
    SignificanceLength: str
    KeepAssume: str
    GlobalsPrefix: str
    UseMacros: str
    StaticLocals: str

    # Test execution fields
    TestHarness: str
    TestResultFile: str

    # Simulation fields
    FileScenario: str
    SimulationInputType: str
    UseCycleTime: str
    CycleTime: str


class _JobData(TypedDict):
    """Job data structure."""

    Version: str
    Kind: str
    Properties: _PropertiesData
    InputPaths: list[str]


class JobStatus(Enum):
    """Exit codes for a job execution"""

    #: Success
    Success = 0
    #: Missing parameters
    MissingParameters = 1
    #: Load issue
    LoadIssue = 2
    #: Not existing project
    NotExistingProject = 3
    #: Job not found
    JobNotFound = 4
    #: Job failure
    JobFailure = 5
    #: Job name duplicate
    JobNameDuplicate = 6
    #: Internal error
    InternalError = 7
    #: Not executed yet
    NotExecutedYet = 8

    @staticmethod
    def get_message(status_code: int) -> str:
        """Get the message corresponding to a status number

        Parameters
        ----------
        code : int
            Status number

        Returns
        -------
        str
            Status message
        """
        if status_code == JobStatus.Success.value:
            return "Job execution successful"
        if status_code == JobStatus.MissingParameters.value:
            return "The arguments are not filled in"
        if status_code == JobStatus.LoadIssue.value:
            return "The project or one of their dependencies cannot be loaded"
        if status_code == JobStatus.NotExistingProject.value:
            return "The project was not found"
        if status_code == JobStatus.JobNotFound.value:
            return (
                "The project has been loaded as expected "
                "but there is no job that matches the specified name"
            )
        if status_code == JobStatus.JobFailure.value:
            return "The job execution has failed"
        if status_code == JobStatus.JobNameDuplicate.value:
            return "There is more than one job matching the requested name"
        if status_code == JobStatus.InternalError.value:
            return "Undefined error"
        if status_code == JobStatus.NotExecutedYet.value:
            return "No job was executed yet"
        return "Unknown error code"

    def __str__(self) -> str:
        return self.get_message(self.value)


class JobResult:
    """Results from the execution of a job"""

    def __init__(self, code: int, message: Optional[str] = None):
        self._code = code
        self._message = message if message else JobStatus.get_message(code)

    @property
    def code(self) -> int:
        """Get execution return code"""
        return self._code

    @property
    def message(self) -> str:
        """Get execution return message"""
        return self._message

    def __str__(self) -> str:
        return f"Code {str(self._code)}: {self._message}"


class ExpansionMode(Enum):
    """Possible values for Expansion field in code generation job properties"""

    #: None
    NONE = auto()
    #: All
    ALL = auto()
    #: Expand
    EXPAND = auto()
    #: NoExpand
    NOEXPAND = auto()

    @staticmethod
    def str_to_expansion(str_exp: Union[str, None]) -> "ExpansionMode":
        """Get Expansion value from sjob file content"""
        if str_exp:
            if str_exp == "All":
                return ExpansionMode.ALL
            if str_exp == "Expand":
                return ExpansionMode.EXPAND
            if str_exp == "NoExpand":
                return ExpansionMode.NOEXPAND
        return ExpansionMode.NONE

    def __str__(self) -> str:
        if self == ExpansionMode.ALL:
            return "All"
        if self == ExpansionMode.EXPAND:
            return "Expand"
        if self == ExpansionMode.NOEXPAND:
            return "NoExpand"
        if self == ExpansionMode.NONE:
            return "None"
        return ""


class JobType(Enum):
    """Job types"""

    #: Code Generation type job
    CODEGEN = auto()
    #: Simulation type job
    SIMU = auto()
    #: Test Execution type job
    TESTEXEC = auto()

    def __str__(self) -> str:
        """String representation of type.
        Used for field `kind` in sjob files"""

        if self == JobType.CODEGEN:
            return "CodeGeneration"
        if self == JobType.SIMU:
            return "Simulation"
        if self == JobType.TESTEXEC:
            return "TestExecution"
        return ""


class Job(ABC):
    """Abstract class for jobs. Must be instantiated from one of its 3 derivated classes.
    Each of them uses its corresponding `Properties` parameters and kind."""

    def __init__(
        self,
        name: str,
        kind: JobType,
        sproj: IProject,
        input_paths: Optional[list[str]] = None,
        storage: JobStorage = None,
    ) -> None:
        self._name = name
        self._kind = kind
        self._version = FormatVersions.version("sjob")
        self._input_paths = input_paths if input_paths else []
        self._storage = storage
        self._sproj = sproj

    def __eq__(self, value: "Job") -> bool:
        """Two jobs are equal if their json content is equal"""
        self.storage.load()
        value.storage.load()
        return self.storage.json == value.storage.json

    @property
    def name(self) -> str:
        """Job name"""
        return self._name

    @property
    def version(self) -> str:
        """.sjob version"""
        return self._version

    @property
    def input_paths(self) -> list[str]:
        """InputPaths parameter"""
        return self._input_paths

    @property
    def sproj(self) -> IProject:
        """Corresponding project"""
        return self._sproj

    @property
    def is_simu(self) -> bool:
        """True if job is a simulation"""
        return self._kind == JobType.SIMU

    @property
    def is_codegen(self) -> bool:
        """True if job is a code generation"""
        return self._kind == JobType.CODEGEN

    @property
    def is_testexec(self) -> bool:
        """True if job is a test execution"""
        return self._kind == JobType.TESTEXEC

    @property
    def storage(self) -> JobStorage:
        """Job storage"""
        return self._storage

    @property
    def root_declarations(self) -> str:
        return self.properties.root_declarations

    @property
    def custom_arguments(self) -> str:
        return self.properties.custom_arguments

    @root_declarations.setter
    def root_declarations(self, value: list[str]) -> None:
        self.properties.root_declarations = value

    @custom_arguments.setter
    def custom_arguments(self, value: str) -> None:
        self.properties.custom_arguments = value

    @name.setter
    def name(self, value: str) -> None:
        self._name = value
        self.properties.name = value

    @property
    @abstractmethod
    def properties(self) -> "JobProperties":
        """Properties parameter.
        Its subtype depends on the Job kind,
        JobProperties is an abstract class."""
        pass

    @input_paths.setter
    def input_paths(self, value: Union[str, list[str]]) -> None:
        """Set InputPaths"""
        if isinstance(value, str):
            self._input_paths = [value]
        else:
            self._input_paths = value

    def folder_name(self) -> str:
        """Generate name of the folder containing the sjob file"""

        if self.is_simu:
            prefix = "simu_"
        elif self.is_codegen:
            prefix = "codegen_"
        elif self.is_testexec:
            prefix = "testexec_"
        else:
            prefix = "unknown_"
        guid_short = str(uuid.uuid4())[:7]
        return prefix + guid_short

    def save(self) -> _JobData:
        """Save the Job parameters into .sjob file.
        Create a new folder for a new Job."""

        data = {}
        data["Version"] = self._version
        data["Kind"] = str(self._kind)
        data["Properties"] = self.properties._to_dict()
        data["InputPaths"] = self._input_paths
        if not self._storage:
            job_folder = self.folder_name()
            sjob_path = Path(self._sproj._storage.source).parent / "jobs" / job_folder / ".sjob"
            self._storage = JobFile(sjob_path)
        parent = Path(self._storage.source).parent
        if not parent.exists():
            import os

            os.makedirs(parent)
        self._storage.json = data
        self._storage.dump()
        return data

    def run(self) -> JobResult:
        job_launcher = JobLauncher()
        job_launcher.job = self
        job_launcher.sproj = self._sproj
        job_launcher.execute()
        return job_launcher.result


class CodeGenerationJob(Job):
    """Job of Code Generation kind"""

    def __init__(
        self, name: str, sproj: IProject, data: dict = None, storage: JobStorage = None
    ) -> None:
        if data:
            super().__init__(name, JobType.CODEGEN, sproj, data.get("InputPaths"), storage)
            self._properties = CodeGenerationJobProperties(name, data.get("Properties"))
        else:
            super().__init__(name, JobType.CODEGEN, sproj)
            self._properties = CodeGenerationJobProperties(name)

    @property
    def properties(self) -> "CodeGenerationJobProperties":
        "Code generation job properties"
        return self._properties

    @property
    def expansion(self) -> ExpansionMode:
        return self.properties.expansion

    @property
    def expansion_exp(self) -> str:
        return self.properties.expansion_exp

    @property
    def expansion_no_exp(self) -> str:
        return self.properties.expansion_no_exp

    @property
    def short_circuit_operators(self) -> bool:
        return self.properties.short_circuit_operators

    @property
    def name_length(self) -> int:
        return self.properties.name_length

    @property
    def significance_length(self) -> int:
        return self.properties.significance_length

    @property
    def keep_assume(self) -> str:
        return self.properties.keep_assume

    @property
    def globals_prefix(self) -> str:
        return self.properties.globals_prefix

    @property
    def use_macros(self) -> bool:
        return self.properties.use_macros

    @property
    def static_locals(self) -> bool:
        return self.properties.static_locals

    @expansion.setter
    def expansion(self, value: Union[ExpansionMode, str, None]) -> None:
        self.properties.expansion = value

    @expansion_exp.setter
    def expansion_exp(self, value: str) -> None:
        self.properties.expansion_exp = value

    @expansion_no_exp.setter
    def expansion_no_exp(self, value: str) -> None:
        self.properties.expansion_no_exp = value

    @short_circuit_operators.setter
    def short_circuit_operators(self, value: bool) -> None:
        self.properties.short_circuit_operators = value

    @name_length.setter
    def name_length(self, value: int) -> None:
        self.properties.name_length = value

    @significance_length.setter
    def significance_length(self, value: int) -> None:
        self.properties.significance_length = value

    @keep_assume.setter
    def keep_assume(self, value: str) -> None:
        self.properties.keep_assume = value

    @globals_prefix.setter
    def globals_prefix(self, value: str) -> None:
        self.properties.globals_prefix = value

    @use_macros.setter
    def use_macros(self, value: bool) -> None:
        self.properties.use_macros = value

    @static_locals.setter
    def static_locals(self, value: bool) -> None:
        self.properties.static_locals = value


class SimulationJob(Job):
    """Job of Simulation kind"""

    def __init__(
        self, name: str, sproj: IProject, data: dict = None, storage: JobStorage = None
    ) -> None:
        if data:
            super().__init__(name, JobType.SIMU, sproj, data.get("InputPaths"), storage)
            self._properties = SimulationJobProperties(name, data.get("Properties"))
        else:
            super().__init__(name, JobType.SIMU, sproj)
            self._properties = SimulationJobProperties(name)

    @property
    def properties(self) -> "SimulationJobProperties":
        """Simulation job properties"""
        return self._properties

    @property
    def file_scenario(self) -> str:
        return self.properties.file_scenario

    @property
    def simulation_input_type(self) -> str:
        return self.properties.simulation_input_type

    @property
    def test_harness(self) -> str:
        return self.properties.test_harness

    @property
    def use_cycle_time(self) -> bool:
        return self.properties.use_cycle_time

    @property
    def cycle_time(self) -> int:
        return self.properties.cycle_time

    @file_scenario.setter
    def file_scenario(self, value: str) -> None:
        self.properties.file_scenario = value

    @simulation_input_type.setter
    def simulation_input_type(self, value: str) -> None:
        self.properties.simulation_input_type = value

    @test_harness.setter
    def test_harness(self, value: str) -> None:
        self.properties.test_harness = value

    @use_cycle_time.setter
    def use_cycle_time(self, value: bool) -> None:
        self.properties.use_cycle_time = value

    @cycle_time.setter
    def cycle_time(self, value: int) -> None:
        self.properties.cycle_time = value


class TestExecutionJob(Job):
    """Job of Test Execution kind"""

    def __init__(
        self, name: str, sproj: IProject, data: dict = None, storage: JobStorage = None
    ) -> None:
        if data:
            super().__init__(name, JobType.TESTEXEC, sproj, data.get("InputPaths"), storage)
            self._properties = TextExecutionJobProperties(name, data.get("Properties"))
        else:
            super().__init__(name, JobType.TESTEXEC, sproj)
            self._properties = TextExecutionJobProperties(name)

    @property
    def properties(self) -> "TextExecutionJobProperties":
        """Test execution job properties"""
        return self._properties

    @property
    def test_harness(self) -> str:
        return self.properties.test_harness

    @property
    def test_result_file(self) -> str:
        return self.properties.test_result_file

    @test_harness.setter
    def test_harness(self, value: str) -> None:
        self.properties.test_harness = value

    @test_result_file.setter
    def test_result_file(self, value: str) -> None:
        self.properties.test_result_file = value


class JobProperties(ABC):
    """Abstract class for job properties, cannot be instantiated.
    Its derived class depends on the kind of Job.
    The parameters presented here can be found in all 3 kinds of Jobs properties.
    """

    def __init__(self, name: str, prop: dict) -> None:
        self._name = name
        self._root_declarations = prop.get("RootDeclarations", [""])
        self._custom_arguments = prop.get("CustomArguments", "")

    @property
    def root_declarations(self) -> list[str]:
        """RootDeclarations common property"""
        return self._root_declarations

    @property
    def custom_arguments(self) -> str:
        """CustomArguments common property"""
        return self._custom_arguments

    @property
    def name(self) -> str:
        """Name common property"""
        return self._name

    @root_declarations.setter
    def root_declarations(self, value: Union[str, list[str]]) -> None:
        """Set RootDeclarations. String can be passed instead of a single element list."""
        if isinstance(value, str):
            self._root_declarations = [value]
        else:
            self._root_declarations = value

    @custom_arguments.setter
    def custom_arguments(self, value: str) -> None:
        """Set CustomArguments"""
        self._custom_arguments = value

    @name.setter
    def name(self, value: str) -> None:
        """Set Name"""
        self._name = value

    def _to_dict(self, data: _PropertiesData) -> _PropertiesData:
        """Return the Properties dictionary that was prefilled
        by one of derivated Properties method"""

        data["RootDeclarations"] = self._root_declarations
        data["CustomArguments"] = self._custom_arguments
        data["Name"] = self._name
        return data


class CodeGenerationJobProperties(JobProperties):
    """Properties of Code Generation kind Jobs.
    Parameters presented here are unique to Code Generation."""

    def __init__(self, name: str, prop: dict = None) -> None:
        if not prop:
            prop = {}
        super().__init__(name, prop)
        self._expansion = ExpansionMode.str_to_expansion(prop.get("Expansion"))
        self._expansion_exp = prop.get("ExpansionExp", "")
        self._expansion_no_exp = prop.get("ExpansionNoExp", "")
        self._short_circuit_operators = prop.get("ShortCircuitOperators") == "True"
        self._name_length = int(prop.get("NameLength", 200))
        self._significance_length = int(prop.get("SignificanceLength", 31))
        self._keep_assume = prop.get("KeepAssume") == "True"
        self._globals_prefix = prop.get("GlobalsPrefix", "")
        self._use_macros = prop.get("UseMacros") == "True"
        self._static_locals = prop.get("StaticLocals") == "True"

    @property
    def expansion(self) -> ExpansionMode:
        """Expansion property"""
        return self._expansion

    @property
    def expansion_exp(self) -> str:
        """ExpansionExp property"""
        return self._expansion_exp

    @property
    def expansion_no_exp(self) -> str:
        """ExpansionNoExp property"""
        return self._expansion_no_exp

    @property
    def short_circuit_operators(self) -> bool:
        """ShortCircuitOperators property"""
        return self._short_circuit_operators

    @property
    def name_length(self) -> int:
        """NameLength property"""
        return self._name_length

    @property
    def significance_length(self) -> int:
        """SignificanceLength property"""
        return self._significance_length

    @property
    def keep_assume(self) -> bool:
        """KeepAssume property"""
        return self._keep_assume

    @property
    def globals_prefix(self) -> str:
        """GlobalsPrefix property"""
        return self._globals_prefix

    @property
    def use_macros(self) -> bool:
        """UseMacros property"""
        return self._use_macros

    @property
    def static_locals(self) -> bool:
        """StaticLocals property"""
        return self._static_locals

    @expansion.setter
    def expansion(self, value: Union[ExpansionMode, str, None]) -> None:
        """Set Expansion"""
        if isinstance(value, ExpansionMode):
            self._expansion = value
        elif isinstance(value, str):
            self._expansion = ExpansionMode.str_to_expansion(value)
        else:
            self._expansion = ExpansionMode.NONE

    @expansion_exp.setter
    def expansion_exp(self, value: str) -> None:
        """Set ExpansionExp"""
        self._expansion_exp = value

    @expansion_no_exp.setter
    def expansion_no_exp(self, value: str) -> None:
        """Set ExpansionNoExp"""
        self._expansion_no_exp = value

    @short_circuit_operators.setter
    def short_circuit_operators(self, value: bool) -> None:
        """Set ShortCircuitOperators"""
        self._short_circuit_operators = value

    @name_length.setter
    def name_length(self, value: int) -> None:
        """Set NameLength"""
        self._name_length = value

    @significance_length.setter
    def significance_length(self, value: int) -> None:
        """Set SignificanceLength"""
        self._significance_length = value

    @keep_assume.setter
    def keep_assume(self, value: bool) -> None:
        """Set KeepAssume"""
        self._keep_assume = value

    @globals_prefix.setter
    def globals_prefix(self, value: str) -> None:
        """Set GlobalsPrefix"""
        self._globals_prefix = value

    @use_macros.setter
    def use_macros(self, value: bool) -> None:
        """Set UseMacros"""
        self._use_macros = value

    @static_locals.setter
    def static_locals(self, value: bool) -> None:
        """Set StaticLocals"""
        self._static_locals = value

    def _to_dict(self) -> _PropertiesData:
        data = {}
        data["Expansion"] = str(self._expansion)
        data["ExpansionExp"] = self._expansion_exp
        data["ExpansionNoExp"] = self._expansion_no_exp
        data["ShortCircuitOperators"] = str(self._short_circuit_operators)
        data["NameLength"] = str(self._name_length)
        data["SignificanceLength"] = str(self._significance_length)
        data["KeepAssume"] = str(self._keep_assume)
        data["GlobalsPrefix"] = self._globals_prefix
        data["UseMacros"] = str(self._use_macros)
        data["StaticLocals"] = str(self._static_locals)
        return super()._to_dict(data)


class SimulationJobProperties(JobProperties):
    """Properties of Simulation kind Jobs.
    Parameters presented here are unique to Simulation."""

    def __init__(self, name: str, prop: dict = None) -> None:
        if not prop:
            prop = {}
        super().__init__(name, prop)
        self._file_scenario = prop.get("FileScenario", "")
        self._simulation_input_type = prop.get("SimulationInputType", "Harness")
        self._test_harness = prop.get("TestHarness", "")  # Optional
        self._use_cycle_time = prop.get("UseCycleTime") == "True"  # bool
        self._cycle_time = int(prop.get("CycleTime", 200))  # Optional int (ms)

    @property
    def file_scenario(self) -> str:
        """FileScenario property"""
        return self._file_scenario

    @property
    def simulation_input_type(self) -> str:
        """SimulationInputType property"""
        return self._simulation_input_type

    @property
    def test_harness(self) -> str:
        """TestHarness property"""
        return self._test_harness

    @property
    def use_cycle_time(self) -> bool:
        """UseCycleTime property"""
        return self._use_cycle_time

    @property
    def cycle_time(self) -> int:
        """CycleTime property"""
        return self._cycle_time

    @file_scenario.setter
    def file_scenario(self, value: str) -> None:
        """Set FileScenario"""
        self._file_scenario = value

    @simulation_input_type.setter
    def simulation_input_type(self, value: str) -> None:
        """Set SimulationInputType"""
        self._simulation_input_type = value

    @test_harness.setter
    def test_harness(self, value: str) -> None:
        """Set TestHarness"""
        self._test_harness = value

    @use_cycle_time.setter
    def use_cycle_time(self, value: bool) -> None:
        """Set UseCycleTime"""
        self._use_cycle_time = value

    @cycle_time.setter
    def cycle_time(self, value: int) -> None:
        """Set CycleTime"""
        self._cycle_time = value

    def _to_dict(self) -> _PropertiesData:
        data = {}
        data["FileScenario"] = self._file_scenario
        data["SimulationInputType"] = self._simulation_input_type
        data["TestHarness"] = self._test_harness
        data["UseCycleTime"] = str(self._use_cycle_time)
        data["CycleTime"] = str(self._cycle_time)
        return super()._to_dict(data)


class TextExecutionJobProperties(JobProperties):
    """Properties of Test Execution kind Jobs.
    Parameters presented here are unique to Test Execution."""

    def __init__(self, name: str, prop: dict = None) -> None:
        if not prop:
            prop = {}
        super().__init__(name, prop)
        self._test_harness = prop.get("TestHarness", "")
        self._test_result_file = prop.get("TestResultFile", "testResults.json")

    @property
    def test_harness(self) -> str:
        """TestHarness property"""
        return self._test_harness

    @property
    def test_result_file(self) -> str:
        """TestResultFile property"""
        return self._test_result_file

    @test_harness.setter
    def test_harness(self, val: str) -> None:
        """Set TestHarness"""
        self._test_harness = val

    @test_result_file.setter
    def test_result_file(self, val: str) -> None:
        """Set TestResultFile"""
        self._test_result_file = val

    def _to_dict(self) -> _PropertiesData:
        data = {}
        data["TestHarness"] = self._test_harness
        data["TestResultFile"] = self._test_result_file
        return super()._to_dict(data)


class JobLauncher:
    """Class used to execute a Job.
    Must set a job and a project in order to execute."""

    def __init__(self) -> None:
        self._sproj = None
        self._job = None
        self._result = JobResult(8)

    @property
    def job(self) -> Union[Job, None]:
        """Job that should be executed"""
        return self._job

    @property
    def sproj(self) -> Union[IProject, None]:
        """Project of the job"""
        return self._sproj

    @property
    def result(self) -> JobResult:
        """Gets the result of the last executed job"""
        return self._result

    @job.setter
    def job(self, value: Job) -> None:
        """Set Job to execute"""
        self._job = value

    @sproj.setter
    def sproj(self, value: IProject) -> None:
        """Set Job's project"""
        self._sproj = value

    def execute(self) -> bool:
        """Execute the Job via JobLauncher executable"""
        from ansys.scadeone.core import ScadeOneException

        if not self._job:
            raise ScadeOneException("No Job was assigned for JobLauncher to execute.")
        elif not self._sproj:
            raise ScadeOneException("No Job project was assigned for the JobLauncher")
        job_launcher = self._sproj.app.tools.job_launcher

        if not job_launcher:
            raise ScadeOneException(
                "Job launcher tool was not found. It is required for Job Execution."
            )

        proc = subprocess.run(
            [str(job_launcher), "run", "-p", self._sproj._storage.source, "-j", self._job.name]
        )
        self._result = JobResult(proc.returncode)
        return self._result.code == JobStatus.Success.value



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-FileCopyrightText: 2024 ANSYS, Inc.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING, List, Union, Optional, cast

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.logger import LOGGER
from ansys.scadeone.core.common.storage import SwanString

if TYPE_CHECKING:
    from ansys.scadeone.core import swan


class DiagramFactory:
    _instance = None

    def __init__(self) -> None:
        from ansys.scadeone.core.model.loader import SwanParser

        self._parser = SwanParser(LOGGER)

    def __new__(cls, *args, **kwargs) -> "DiagramFactory":
        if not cls._instance:
            cls._instance = super(DiagramFactory, cls).__new__(cls)
        return cls._instance

    @staticmethod
    def create_block(diagram: "swan.Diagram", instance: "swan.Operator") -> "swan.Block":
        """Create a call operator block.

        Parameters
        ----------
        diagram: Diagram
            Diagram where the block will be added.
        instance: Operator
            Operator instance. A **use** directive may be required in the *diagram*'s module to
            resolve the operator module.

        Returns
        -------
        Block
            Block object

        Raises
        ------
        ScadeOneException
            If the operator does not have a module or if the diagram does not have a module.
        """
        from ansys.scadeone.core.swan import Block, PathIdOpCall
        from ansys.scadeone.core.model import Model

        path_id = Model.get_path_in_module(instance, diagram.module)
        path_id_op_call = PathIdOpCall(path_id, [], [])
        return Block(path_id_op_call)

    @staticmethod
    def create_def_block(instance: "swan.VarDecl") -> "swan.DefBlock":
        """Create a definition block from a variable declaration."""
        from ansys.scadeone.core.swan import DefBlock, EquationLHS, LHSItem

        item = LHSItem(instance.id)
        eq = EquationLHS([item])
        return DefBlock(eq)

    @staticmethod
    def create_expr_block(
        diagram: "swan.Diagram", declaration: "swan.Declaration"
    ) -> "swan.ExprBlock":
        """Create an expression block from a variable declaration."""
        from ansys.scadeone.core.swan import ExprBlock, PathIdExpr, ConstDecl, VarDecl, SensorDecl
        from ansys.scadeone.core.model import Model

        if not isinstance(declaration, (ConstDecl, VarDecl, SensorDecl)):
            raise ScadeOneException("declaration must be a constant, variable or sensor object")
        path_id = Model.get_path_in_module(declaration, diagram.module)
        path_id_expr = PathIdExpr(path_id)
        return ExprBlock(path_id_expr)

    @staticmethod
    def create_bar() -> "swan.Bar":
        """Create a group."""
        from ansys.scadeone.core.swan import Bar

        return Bar()

    def create_wire(
        self,
        source: Union["swan.DiagramObject", tuple["swan.DiagramObject", str]],
        targets: Union[List["swan.DiagramObject"], List[tuple["swan.DiagramObject", str]]],
    ) -> "swan.Wire":
        """Create a wire between a source block and one or more targets block
        with or without group declarations."""
        from ansys.scadeone.core.swan import Connection, PortExpr, Wire

        source_obj, source_adp = self._split_wire_end(source)
        source_port_expr = PortExpr(source_obj.lunum)
        source_conn = Connection(source_port_expr, source_adp)
        target_conns = []
        for target in targets:
            target_obj, target_adp = self._split_wire_end(target)
            target_port_expr = PortExpr(target_obj.lunum)
            target_conns.append(Connection(target_port_expr, target_adp))
        return Wire(source_conn, target_conns)

    def _split_wire_end(
        self, end: Union["swan.DiagramObject", tuple["swan.DiagramObject", str]]
    ) -> tuple["swan.DiagramObject", "swan.GroupAdaptation"]:
        """Return a tuple containing the diagram object and the group adaptation
        of a wire source or target"""
        if isinstance(end, tuple):
            end_obj = end[0]
            end_adp = self._parse_group_adaptation(end[1])
        else:
            end_obj = end
            end_adp = None
        return end_obj, end_adp

    def _parse_group_adaptation(self, adp: str) -> "swan.GroupAdaptation":
        """Convert a group adaptation expression into a GroupAdaptation object."""
        from ansys.scadeone.core.swan import Diagram, Wire

        swan_code = f"diagram (#0 wire #1 .({adp}) => #2)"
        diag_str = SwanString(swan_code, "new_op")
        diag = cast(Diagram, self._parser.scope_section(diag_str))
        wire = cast(Wire, diag.objects[0])
        return wire.source.adaptation


class DiagramAdder:
    def __init__(self, owner: "swan.Diagram") -> None:
        self._owner = owner
        self._lunum = -1

    def add_block(self, block: "swan.Block") -> None:
        """Add a block to the diagram."""
        self._add_diagram_object(block)

    def add_def_block(self, def_block: "swan.DefBlock") -> None:
        """Add a definition block to the diagram."""
        self._add_diagram_object(def_block)

    def add_expr_block(self, expr_block: "swan.ExprBlock") -> None:
        """Add an expression block to the diagram."""
        self._add_diagram_object(expr_block)

    def add_bar(self, bar: "swan.Bar") -> None:
        """Add a group to the diagram."""
        self._add_diagram_object(bar)

    def _check_for_wire(self, wire: "swan.Wire") -> Optional["swan.Wire"]:
        # Check if wire is already in the diagram with the same source characteristics
        from ansys.scadeone.core.swan import Wire

        source_lunum = wire.source.port.lunum.value
        source_adaptation = wire.source.adaptation
        # look for candidate wires with the same source lunum
        candidates = [obj for obj in self._owner.objects if isinstance(obj, Wire)]

        def _check(w: "swan.Wire") -> bool:
            same_lunum = source_lunum == w.source.port.lunum.value
            same_adaptation = source_adaptation == w.source.adaptation
            return same_lunum and same_adaptation

        wires = list(filter(_check, candidates))
        if len(wires) == 0:
            return None
        return wires[0]

    def add_wire(self, wire: "swan.Wire") -> "swan.Wire":
        """Add a wire to the diagram."""
        if existing_wire := self._check_for_wire(wire):
            # Check if the wire is already in the diagram with the same source
            # and add the target to the existing wire
            existing_wire.targets.extend(wire.targets)
            return existing_wire

        # No candidate wires found, add the wire to the diagram
        self._add_diagram_object(wire)
        return wire

    def _add_diagram_object(self, object: "swan.DiagramObject") -> None:
        """Add a diagram object to the diagram."""
        from ansys.scadeone.core.swan import Lunum

        self._generate_next_lunum()
        object._lunum = Lunum(f"#{self._lunum}")
        self._lunum += 1
        object.owner = self._owner
        if not self._owner._objects:
            self._owner._objects = [object]
        else:
            self._owner._objects.append(object)

    def _generate_next_lunum(self) -> None:
        """Generate the next lunum for the diagram objects."""
        if not self._lunum == -1:
            return
        if not self._owner._objects:
            self._lunum = 0
        else:
            self._lunum = max([int(obj.lunum.value[1:]) for obj in self._owner._objects]) + 1


class DiagramCreator:
    def __init__(self) -> None:
        self._diagram_adder = DiagramAdder(self)

    def add_block(self, instance: "swan.Operator") -> "swan.Block":
        """Add an operator call to the operator.

        Parameters
        ----------
        instance: Operator
            Operator instance.

        Returns
        -------
        Block
            Block object.
        """
        block = DiagramFactory().create_block(self, instance)
        self._diagram_adder.add_block(block)
        return block

    def add_def_block(self, variable: "swan.VarDecl") -> "swan.DefBlock":
        """Add a definition block to the operator.

        Parameters
        ----------
        variable: VarDecl
            Local, input or output variable.

        Returns
        -------
        DefBlock
            DefBlock object.
        """
        def_block = DiagramFactory().create_def_block(variable)
        self._diagram_adder.add_def_block(def_block)
        return def_block

    def add_expr_block(self, declaration: "swan.Declaration") -> "swan.ExprBlock":
        """Add an expression block to the operator.

        Parameters
        ----------
        declaration: Declaration
            Local or output variable.

        Returns
        -------
        ExprBlock
            ExprBlock object.
        """
        expr_block = DiagramFactory().create_expr_block(self, declaration)
        self._diagram_adder.add_expr_block(expr_block)
        return expr_block

    def add_bar(self) -> "swan.Bar":
        """Add a bar to the operator.

        Returns
        -------
            Bar
                Bar object.
        """
        bar = DiagramFactory().create_bar()
        self._diagram_adder.add_bar(bar)
        return bar

    def connect(
        self,
        source: Union["swan.DiagramObject", tuple["swan.DiagramObject", str]],
        targets: Union[
            "swan.DiagramObject",
            tuple["swan.DiagramObject", str],
            List["swan.DiagramObject"],
            List[tuple["swan.DiagramObject", str]],
        ],
    ) -> "swan.Wire":
        """Connect a block to one or more blocks.

        Parameters
        ----------
        source: DiagramObject
            Block object.
        targets: Union[DiagramObject, List[DiagramObject]]
            A block object or a list of block objects.

        Returns
        -------
        Wire
            Wire object that connects the blocks.
        """
        if not isinstance(targets, list):
            targets = [targets]
        wire = DiagramFactory().create_wire(source, targets)
        added_wire = self._diagram_adder.add_wire(wire)
        return added_wire



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-FileCopyrightText: 2024 ANSYS, Inc.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import ABC
from typing import TYPE_CHECKING, List, Optional, Union, cast

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.logger import LOGGER
from ansys.scadeone.core.common.storage import SwanString

if TYPE_CHECKING:
    import ansys.scadeone.core.swan as swan


class OperatorFactory:
    _instance = None

    def __init__(self) -> None:
        from ansys.scadeone.core.model.loader import SwanParser

        self._parser = SwanParser(LOGGER)

    def __new__(cls, *args, **kwargs) -> "OperatorFactory":
        if not cls._instance:
            cls._instance = super(OperatorFactory, cls).__new__(cls)
        return cls._instance

    def create_variable(
        self,
        name: Optional[str] = None,
        var_type: Union[str, "swan.Declaration"] = None,
        is_clock: Optional[bool] = False,
        is_probe: Optional[bool] = False,
        when: Optional[str] = None,
        default: Optional[str] = None,
        last: Optional[str] = None,
        declaration: Optional[str] = None,
        is_input: Optional[bool] = None,
        module: Optional["swan.Module"] = None,
    ) -> "swan.VarDecl":
        """Create a variable.
        A variable can be defined by name, type, ... or by a text declaration."""
        from ansys.scadeone.core.swan import (
            Declaration,
            Identifier,
            Signature,
            VarDecl,
            PathIdentifier,
        )
        from ansys.scadeone.core.model import Model

        if (not name or not var_type) and not declaration:
            raise ScadeOneException(
                "Invalid variable parameters: declaration or name and type are required"
            )
        if isinstance(var_type, Declaration):
            if module is None:
                raise ScadeOneException("Module is required for OperatorFactor.create_variable()")
            type_path: PathIdentifier = Model.get_path_in_module(var_type, module)
            var_type = type_path.as_string
        if name and var_type:
            id = Identifier(name)
            if not id.is_valid:
                raise ScadeOneException(f"Invalid variable name: {name}")
            declaration = f"{name}: {var_type}"
            if when:
                declaration += f" when {when}"
            if default:
                declaration += f" default = {default}"
            if last:
                declaration += f" last = {last}"
        if not declaration:
            raise ScadeOneException("Invalid variable declaration")
        if is_input is False:
            swan_code = f"function temp_op(i0: int32) returns ({declaration});"
        else:
            swan_code = f"function temp_op({declaration}) returns (o0: int32);"
        op_str = SwanString(swan_code, "new_op")
        op = cast(Signature, self._parser.operator_decl(op_str))
        if not op:
            raise ScadeOneException("Invalid operator declaration")
        if is_input is False:
            var = cast(VarDecl, op.outputs[0])
        else:
            var = cast(VarDecl, op.inputs[0])
        var._is_clock = is_clock
        var._is_probe = is_probe
        return var

    @staticmethod
    def create_signature(
        name: Union[List["swan.Identifier"], str],
        is_node: bool = True,
        is_inline: bool = False,
    ) -> "swan.Signature":
        """Create a signature."""
        from ansys.scadeone.core.swan import Identifier, Signature

        id = Identifier(name)
        if not id.is_valid:
            raise ScadeOneException(f"Invalid module name: {name}")
        return Signature(id, is_inline, is_node, [], [])

    @staticmethod
    def create_operator(
        name: Union[List["swan.Identifier"], str],
        is_node: bool = True,
        is_inline: bool = False,
    ) -> "swan.Operator":
        """Create an operator."""
        from ansys.scadeone.core.swan import Identifier, Operator

        id = Identifier(name)  # noqa
        if not id.is_valid:
            raise ScadeOneException(f"Invalid module name: {name}")
        return Operator(id, is_inline, is_node, [], [])


class OperatorAdder:
    @staticmethod
    def add_input(operator: "swan.Signature", variable: "swan.Variable") -> None:
        """Add an input to the operator."""
        variable._is_input = True
        operator.inputs.append(variable)
        variable._owner = operator

    @staticmethod
    def add_output(operator: "swan.Signature", variable: "swan.Variable") -> None:
        """Add an output to the operator."""
        variable._is_output = True
        operator.outputs.append(variable)
        variable._owner = operator


class SignatureCreator(ABC):
    def _add_variable(
        self,
        name: Optional[str] = None,
        var_type: Union[str, "swan.Declaration"] = None,
        is_clock: Optional[bool] = False,
        is_probe: Optional[bool] = False,
        when: Optional[str] = None,
        default: Optional[str] = None,
        last: Optional[str] = None,
        declaration: Optional[str] = None,
        is_input: Optional[bool] = None,
    ) -> "swan.VarDecl":
        """Add an input to the operator.
        Input can be defined by name, type, ... or by a text declaration.

        Parameters
        ----------
        name: str
            Variable name.
        var_type: str or Declaration
            Variable type.
            It could be a text declaration or a Declaration object.
        is_clock: bool
            Clock variable.
        is_probe: bool
            Probe variable.
        when: str
            When condition.
        default: str
            Default value.
        last: str
            Last value.
        declaration: str
            Text declaration.

        Returns
        -------
        VarDecl
            Variable object.
        """
        from ansys.scadeone.core.swan.operators import OperatorSignatureBase

        if not isinstance(self, OperatorSignatureBase):
            raise ScadeOneException("SignatureCreator: expecting Operator or Signature object")
        var = OperatorFactory().create_variable(
            name=name,
            var_type=var_type,
            is_clock=is_clock,
            is_probe=is_probe,
            when=when,
            default=default,
            last=last,
            declaration=declaration,
            is_input=is_input,
            module=None if isinstance(var_type, str) else self.module,
        )
        return var

    def add_input(
        self,
        name: Optional[str] = None,
        input_type: Union[str, "swan.Declaration"] = None,
        is_clock: Optional[bool] = False,
        is_probe: Optional[bool] = False,
        when: Optional[str] = None,
        default: Optional[str] = None,
        last: Optional[str] = None,
        declaration: Optional[str] = None,
    ) -> "swan.VarDecl":
        """Add an input to the operator.
        Input can be defined by name, type, ... or by a text declaration.

        Parameters
        ----------
        name: str
            Input name.
        input type: str or Declaration
            Input type.
            It could be a text declaration or a Declaration object.
        is_clock: bool
            Clock variable.
        is_probe: bool
            Probe variable.
        when: str
            When condition.
        default: str
            Default value.
        last: str
            Last value.
        declaration: str
            Text declaration.

        Returns
        -------
        VarDecl
            Variable object.
        """
        var = self._add_variable(
            name=name,
            var_type=input_type,
            is_clock=is_clock,
            is_probe=is_probe,
            when=when,
            default=default,
            last=last,
            declaration=declaration,
            is_input=True,
        )
        OperatorAdder.add_input(self, var)
        return var

    def add_output(
        self,
        name: Optional[str] = None,
        output_type: Union[str, "swan.Declaration"] = None,
        is_clock: Optional[bool] = False,
        is_probe: Optional[bool] = False,
        when: Optional[str] = None,
        default: Optional[str] = None,
        last: Optional[str] = None,
        declaration: Optional[str] = None,
    ) -> "swan.VarDecl":
        """Add an output to the operator signature.
        Output can be defined by name, type, ... or by a text declaration.

        Parameters
        ----------
        name: str
            Variable name.
        output_type: str or Declaration
            Output type.
            It could be a text declaration or a Declaration object.
        is_clock: bool
            Clock variable.
        is_probe: bool
            Probe variable.
        when: str
            When condition.
        default: str
            Default value.
        last: str
            Last value.
        declaration: str
            Textual declaration.

        Returns
        -------
        VarDecl
            Variable object.
        """
        var = self._add_variable(
            name=name,
            var_type=output_type,
            is_clock=is_clock,
            is_probe=is_probe,
            when=when,
            default=default,
            last=last,
            declaration=declaration,
            is_input=False,
        )
        OperatorAdder.add_output(self, var)
        return var


class OperatorCreator(SignatureCreator):
    def add_diagram(self) -> "swan.Diagram":
        """Add a diagram to the operator.

        Returns
        -------
        Diagram
            Diagram object.
        """
        from ansys.scadeone.core.swan import Diagram, Operator, Scope

        if not isinstance(self, Operator):
            raise ScadeOneException("OperatorCreator must be used with an Operator object")
        diag = Diagram()
        if not self.body or not self.body.sections:
            scope = Scope([diag])
            scope.owner = self
            self._body = scope
            return diag
        self._body._sections.append(diag)
        return diag



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This module contains the classes and functions that are necessary to
# manipulate Scade One model elements retrieved from the mapping file.

# flake8: noqa: E101

from abc import ABC, abstractmethod
from enum import Enum, auto
import functools

# cSpell: ignore oper, elems, ename
import json
from pathlib import Path
from typing import Generator, List, Optional, Tuple, Union

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.project import Project
from ansys.scadeone.core.common.exception import LOGGER

MapElem = List[Union[str, dict]]
CodeElem = Tuple[str, dict, int]


class MappingObject(ABC):
    """
    Base class for all mapping file elements.

    - gen_code: *GeneratedCode* object
    - mapping_elem: mapping file element
    """

    def __init__(self, gen_code: "GeneratedCode", mapping_elem: dict) -> None:
        self._id = mapping_elem["id"]
        self._gen_code = gen_code
        self._mapping_elem = mapping_elem


class CDeclaration(MappingObject):
    """
    Base class for C declaration
    (function, global, macro, predefined_type, struct, array, enum, union, typedef, imported_type).

    - gen_code: *GeneratedCode* object
    - decl_elem: mapping file element
    - code_container_index: position in the list of code container elements
    """

    def __init__(
        self, gen_code: "GeneratedCode", decl_elem: dict, code_container_index: int
    ) -> None:
        super().__init__(gen_code, decl_elem)
        self._code_container_index = code_container_index

    def get_interface_file(self) -> str:
        """Returns the name of the enclosed interface C header file."""
        try:
            return self._gen_code.code[self._code_container_index]["interface_file"]
        except Exception:
            raise ScadeOneException(
                f"Generated code: no interface_file for CodeContainer at index #{self._code_container_index}"  # noqa
            )

    def get_implementation_file(self) -> str:
        """Returns the name of the enclosed implementation C file."""
        try:
            return self._gen_code.code[self._code_container_index]["implementation_file"]
        except Exception:
            raise ScadeOneException(
                f"Generated code: no implementation_file for CodeContainer at index #{self._code_container_index}"  # noqa
            )


class CParameter(MappingObject):
    """
    Class for a parameter of a generated C function.

    - gen_code: *GeneratedCode* object
    - param_elem: mapping file element
    """

    def __init__(self, gen_code: "GeneratedCode", param_elem: dict) -> None:
        super().__init__(gen_code, param_elem)
        self._name = param_elem["name"]
        self._type_id = param_elem["type"]
        self._pointer = param_elem.get("pointer", False)
        self._const = param_elem.get("const", False)
        self._type_name = None

    @property
    def param_type_id(self) -> int:
        """Returns the type id of the parameter"""
        return self._type_id

    @property
    def name(self) -> str:
        """Returns the parameter name"""
        return self._name

    @property
    def type_name(self) -> str:
        """Returns the parameter type name"""
        if self._type_name is None:
            type_elem = self._gen_code.get_code_elem(self._type_id)
            self._type_name = type_elem.get("content", {}).get("name", "__unknown__")
        return self._type_name

    @property
    def pointer(self) -> bool:
        """Indicates if the parameter is a pointer"""
        return self._pointer

    @property
    def const(self) -> bool:
        """Indicates if the parameter is a const type"""
        return self._const

    @property
    def signature(self) -> str:
        """Returns the parameter type signature (taking into account pointer and const)."""
        const = "const " if self.const else ""
        pointer = " *" if self.pointer else ""
        return f"{const}{self.type_name}{pointer}"


class CFunction(CDeclaration):
    """
    Class for a generated C function.

    - gen_code: *GeneratedCode* object
    - function_elem: mapping file element
    - code_container_index: position in the list of code container elements
    """

    def __init__(
        self, gen_code: "GeneratedCode", function_elem: dict, code_container_index: int
    ) -> None:
        super().__init__(gen_code, function_elem, code_container_index)
        self._name = None
        self._parameters = None
        self._return_type = None

    @property
    def name(self) -> str:
        """Returns the function name"""
        if self._name is None:
            self._name = self._mapping_elem["name"]
        return self._name

    @property
    def parameters(self) -> List[CParameter]:
        """Returns the function parameters as a list of CParameter objects."""
        if self._parameters is None:
            self._parameters = [
                CParameter(self._gen_code, x) for x in self._mapping_elem.get("parameters", [])
            ]
        return self._parameters

    @property
    def return_type(self) -> Optional[dict]:
        """Returns the function return type as raw mapping file dictionary data, if any."""
        if self._return_type is None:
            code_id = self._mapping_elem.get("return_type", -1)
            if code_id != -1:
                self._return_type = self._gen_code.decompose_code_type(
                    self._gen_code.get_code_elem(code_id)
                )
        return self._return_type


class ModelObject(MappingObject):
    """
    Base Class for mapping file model elements.

    - gen_code: *GeneratedCode* object
    - model_elem: mapping file element
    """

    def __init__(self, gen_code: "GeneratedCode", model_elem: dict) -> None:
        super().__init__(gen_code, model_elem)

    def get_code_elem(self, role: str = "", silent: bool = False) -> Optional[dict]:
        """Returns the data associated to the model element, if any."""
        code_id = self._gen_code.get_code_id(self._id, role, silent)
        if code_id != -1:
            return self._gen_code.get_code_elem(code_id)
        else:
            return None


class ModelSensor(ModelObject):
    """
    Class for model sensors.

    - gen_code: *GeneratedCode* object
    - var_elem: mapping file element
    """

    def __init__(self, gen_code: "GeneratedCode", sensor_elem: dict) -> None:
        super().__init__(gen_code, sensor_elem)
        self._path = None
        self._type_id = sensor_elem.get("type", -1)
        self._code_name = None
        self._code_type = None
        self._code_type_name = None

    @property
    def path(self) -> str:
        """Returns the Scade One sensor path"""
        if self._path is None:
            self._path = self._mapping_elem["path"]
        return self._path

    @property
    def code_name(self) -> str:
        """Returns the name of the variable as defined in the generated C function."""
        if self._code_name is None:
            code_elem = self.get_code_elem(silent=True)
            if code_elem is None:
                self._code_name = "__no_name__"
            else:
                self._code_name = code_elem["content"]["name"]
        return self._code_name

    @property
    def code_type(self) -> dict:
        """
        Returns the generated C code data corresponding to the sensor type.
        ('name': C type name, 'category': type category and 'elements': sub elements of the type)
        """
        if self._code_type is None:
            code_elem = self.get_code_elem(silent=True)
            self._code_type = self._gen_code.decompose_code_type(
                self._gen_code.get_code_elem(code_elem["content"]["type"])
            )
        return self._code_type

    @property
    def code_type_name(self) -> str:
        """Returns the sensor C type name"""
        if self._code_type_name is None:
            self._code_type_name = self.code_type.get("name", "__unknown__")
        return self._code_type_name

    def get_code_elem(self, role: str = "", silent: bool = False) -> Optional[dict]:
        """Returns the data associated to the model element, if any."""
        code_elem = super().get_code_elem(role, True)
        if code_elem is None:
            # no correspondence in mapping section => use the model id as the code id
            code_elem = self._gen_code.get_code_elem(self._id)
        return code_elem


class ModelVariableBase(ModelObject):
    """
    Base class for model variables (operators inputs or outputs).

    - gen_code: *GeneratedCode* object
    - var_elem: mapping file element
    """

    def __init__(self, gen_code: "GeneratedCode", var_elem: dict) -> None:
        super().__init__(gen_code, var_elem)
        self._code_name = None
        self._code_type = None
        self._code_elem = None

    @property
    @abstractmethod
    def parent(self) -> "ModelOperatorBase":
        """Returns the model operator object owning the variable"""
        pass

    @property
    def code_name(self) -> str:
        """Returns the name of the variable as defined in the generated C function."""
        if self._code_name is None:
            if self._code_elem is None:
                self._code_elem = self.get_code_elem(silent=True)
            if self._code_elem is None:
                self._code_name = "__no_name__"
            elif (
                self._code_elem["category"] == "fields"
                and self._code_elem["parent_category"] == "struct"
            ):
                # parameter is defined in struct => find the name of the corresponding parameter
                # in the function
                func = self.parent.cycle_method
                ctx_path = "__unknown__"
                for p in func.parameters:
                    if p.param_type_id == self._code_elem["parent_id"]:
                        ctx_path = p.name
                        break
                self._code_name = ctx_path + "." + self._code_elem["content"]["name"]
            else:
                self._code_name = self._code_elem["content"]["name"]
        return self._code_name

    @property
    def code_type(self) -> dict:
        """
        Returns the generated C code data corresponding to the variable type.
        ('name': C type name, 'category': type category and 'elements': sub elements of the type)
        """
        if self._code_type is None:
            if self._code_elem is None:
                self._code_elem = self.get_code_elem(silent=True)
            if self._code_elem is None:
                # no code associated to variable => it is the return
                # of the cycle function or a polymorphic param
                if self.parent.cycle_method is None:
                    self._code_type = {}
                else:
                    self._code_type = self.parent.cycle_method.return_type
                    if self._code_type is None:
                        raise ScadeOneException(
                            f"Generated code: return_type missing for CycleMethod of operator {self.parent.path}"  # noqa
                        )
            else:
                self._code_type = self._gen_code.decompose_code_type(
                    self._gen_code.get_code_elem(self._code_elem["content"]["type"])
                )
        return self._code_type


class ModelVariable(ModelVariableBase):
    """
    Class for model variables (inputs or outputs) of operator.

    - gen_code: *GeneratedCode* object
    - var_elem: mapping file element
    - parent: *ModelOperator* object of the parent operator
    """

    def __init__(self, gen_code: "GeneratedCode", var_elem: dict, parent: "ModelOperator") -> None:
        super().__init__(gen_code, var_elem)
        self._parent = parent
        self._name = var_elem["name"]
        self._projection = var_elem.get("projection", "")

    @property
    def parent(self) -> "ModelOperator":
        """Returns the model operator object owning the variable"""
        return self._parent

    @property
    def name(self) -> str:
        """Returns the variable name"""
        return self._name

    @property
    def group_items(self) -> List[Union[str, int]]:
        """Returns the group item(s) associated with variable (only applicable for group type)."""
        return self._projection

    def full_name(self, separator: str = "_") -> str:
        """
        Returns variable name taking into account group item(s) to suffix the name when applicable
        (main separator between the operator variable name and group item(s) name can be changed,
        default is '_')
        """
        group_suffix = "_".join("_" + str(i) if isinstance(i, int) else i for i in self.group_items)
        return self.name + (separator if len(self.group_items) > 0 else "") + group_suffix


class ModelVariableMonomorphic(ModelVariableBase):
    """
    Class for model variables (inputs or outputs) of monomorphic instance of operator.

    - gen_code: *GeneratedCode* object
    - var_elem: mapping file element
    - parent: *ModelMonomorphicInstance* object of the parent monomorphic instance
    """

    def __init__(
        self,
        gen_code: "GeneratedCode",
        var_elem: dict,
        parent: "ModelMonomorphicInstance",
    ) -> None:
        super().__init__(gen_code, var_elem)
        self._parent = parent
        self._source = None

    @property
    def parent(self) -> "ModelMonomorphicInstance":
        """Returns the model operator object owning the variable"""
        return self._parent

    @property
    def source(self) -> "ModelVariable":
        """Returns ModelVariable object source of the monomorphic instance variable."""
        if self._source is None:
            # get id of corresponding model variable
            id = self._mapping_elem.get("src", 0)
            # browse the inputs and outputs of the operator source of the instance
            for i in self.parent.source.inputs:
                if i._id == id:
                    self._source = i
                    break
            if self._source is None:
                for i in self.parent.source.outputs:
                    if i._id == id:
                        self._source = i
                        break
        return self._source


class ModelOperatorBase(ModelObject):
    """
    Base class for model operator or monomorphic instance of an operator.

    - gen_code: *GeneratedCode* object
    - oper_elem: mapping file element
    """

    def __init__(self, gen_code: "GeneratedCode", oper_elem: dict) -> None:
        super().__init__(gen_code, oper_elem)
        self._path = None
        self._inputs = None
        self._outputs = None
        self._watches = None
        self._instances = None
        self._cycle_method = None
        self._init_method = None
        self._reset_method = None

    @property
    def path(self) -> str:
        """Returns the Scade One operator path"""
        if self._path is None:
            self._path = self._mapping_elem["path"]
        return self._path

    @property
    def watches(self) -> List[dict]:
        """Returns the list of watches for the operator (as raw data, to be completed)."""
        if self._watches is None:
            self._watches = self._mapping_elem.get("watches", [])
        return self._watches

    @property
    def instances(self) -> List[dict]:
        """Returns the list of instances for the operator (as raw data, to be completed)."""
        if self._instances is None:
            self._instances = self._mapping_elem.get("instances", [])
        return self._instances

    @property
    def cycle_method(self) -> Optional["CFunction"]:
        """Returns the cycle method C function for the operator (as a CFunction object), if any."""
        if self._cycle_method is None:
            method_elem = self.get_code_elem("CycleMethod", silent=True)
            if method_elem is not None:
                self._cycle_method = CFunction(
                    self._gen_code, method_elem["content"], method_elem["code_index"]
                )
        return self._cycle_method

    @property
    def init_method(self) -> Optional["CFunction"]:
        """Returns the init method C function for the operator (as a CFunction object), if any."""
        if self._init_method is None:
            method_elem = self.get_code_elem("InitMethod", silent=True)
            if method_elem is not None:
                self._init_method = CFunction(
                    self._gen_code, method_elem["content"], method_elem["code_index"]
                )
        return self._init_method

    @property
    def reset_method(self) -> Optional["CFunction"]:
        """Returns the reset method C function for the operator (as a CFunction object), if any."""
        if self._reset_method is None:
            method_elem = self.get_code_elem("ResetMethod", silent=True)
            if method_elem is not None:
                self._reset_method = CFunction(
                    self._gen_code, method_elem["content"], method_elem["code_index"]
                )
        return self._reset_method


class ModelOperator(ModelOperatorBase):
    """
    Class for model operator.

    - gen_code: *GeneratedCode* object
    - oper_elem: mapping file element
    """

    def __init__(self, gen_code: "GeneratedCode", oper_elem: dict) -> None:
        super().__init__(gen_code, oper_elem)
        self._root = oper_elem.get("root", False)
        self._imported = oper_elem.get("imported", False)
        self._expanded = oper_elem.get("expanded", False)
        self._specialize = oper_elem.get("specialize", 0)

    @property
    def root(self) -> bool:
        """Returns True if the operator is a root operator"""
        return self._root

    @property
    def imported(self) -> bool:
        """Returns True if the operator is an imported operator"""
        return self._imported

    @property
    def expanded(self) -> bool:
        """Returns True if the operator is expanded (inlined)"""
        return self._expanded

    @property
    def specialize(self) -> bool:
        """Returns True if the operator is specialized"""
        return self._specialize

    @property
    def inputs(self) -> List[ModelVariable]:
        """Returns the list of inputs for the operator (as ModelVariable objects)."""
        if self._inputs is None:
            self._inputs = [
                ModelVariable(self._gen_code, x, self) for x in self._mapping_elem.get("inputs", [])
            ]
        return self._inputs

    @property
    def outputs(self) -> List[ModelVariable]:
        """Returns the list of outputs for the operator (as ModelVariable objects)."""
        if self._outputs is None:
            self._outputs = [
                ModelVariable(self._gen_code, x, self)
                for x in self._mapping_elem.get("outputs", [])
            ]
        return self._outputs


class ModelMonomorphicInstance(ModelOperatorBase):
    """
    Class for monomorphic instance of polymorphic model operator.

    - gen_code: *GeneratedCode* object
    - oper_elem: mapping file element
    """

    def __init__(self, gen_code: "GeneratedCode", oper_elem: dict) -> None:
        super().__init__(gen_code, oper_elem)
        self._source = None
        self._type_parameters = oper_elem.get("type_parameters", [])
        self._size_parameters = oper_elem.get("size_parameters", [])

    @property
    def source(self) -> "ModelOperator":
        """Returns ModelOperator object source of the monomorphic instance."""
        if self._source is None:
            id = self._mapping_elem.get("src", 0)
            try:
                _, op_elem = next(
                    i for i in self._gen_code.get_model_elements("operator") if i[1]["id"] == id
                )
                self._source = ModelOperator(self._gen_code, op_elem)
            except StopIteration:
                raise ScadeOneException(
                    f"Generated code: can't find source operator (id: {id}) for {self.path}"
                )
        return self._source

    @property
    def inputs(self) -> List[ModelVariableMonomorphic]:
        """Returns the list of inputs for the operator (as ModelVariableMonomorphic objects)."""
        if self._inputs is None:
            self._inputs = [
                ModelVariableMonomorphic(self._gen_code, x, self)
                for x in self._mapping_elem.get("inputs", [])
            ]
        return self._inputs

    @property
    def outputs(self) -> List[ModelVariableMonomorphic]:
        """Returns the list of outputs for the operator (as ModelVariableMonomorphic objects)."""
        if self._outputs is None:
            self._outputs = [
                ModelVariableMonomorphic(self._gen_code, x, self)
                for x in self._mapping_elem.get("outputs", [])
            ]
        return self._outputs


class ModelKeyword(Enum):
    Array = auto()
    Enum = auto()
    EnumValue = auto()
    Field = auto()
    Function = auto()
    Global = auto()
    Inputs = auto()
    Operator = auto()
    Outputs = auto()
    PredefinedType = auto()
    Sensor = auto()
    Struct = auto()

    @staticmethod
    def to_str(value: "ModelKeyword") -> str:
        if value == ModelKeyword.PredefinedType:
            return "predefined_type"
        elif value == ModelKeyword.EnumValue:
            return "enum_value"
        else:
            return value.name.lower()


class TypeObject(ABC):
    def __init__(self, c_name: str = "", m_name: str = "", path: str = ""):
        self.c_name: str = c_name
        self.m_name: str = m_name
        self.path: str = path


class TypeBase(TypeObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def is_scalar(self) -> bool:
        return isinstance(self, Scalar)

    @property
    def is_array(self) -> bool:
        return isinstance(self, Array)

    @property
    def is_structure(self) -> bool:
        return isinstance(self, Structure)

    @property
    def is_enum_type(self) -> bool:
        return isinstance(self, Enumerate)


class Scalar(TypeBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Enumerate(TypeBase):
    def __init__(self, enum_name, *args, **kwargs):
        self._enum_name = enum_name
        super().__init__(*args, **kwargs)

    @property
    def enum_name(self) -> str:
        return self._enum_name


class Array(TypeBase):
    def __init__(self, sizes: list[int], *args, **kwargs):
        self._sizes = sizes
        super().__init__(*args, **kwargs)

    @property
    def sizes(self) -> list[int]:
        return self._sizes


class Structure(TypeBase):
    def __init__(self, fields: list[TypeBase], *args, **kwargs):
        self._fields = fields
        super().__init__(*args, **kwargs)

    @property
    def fields(self) -> list[TypeBase]:
        return self._fields


class GeneratedCode(object):
    """
    Generated code main class.

    Parameters
    ----------
    project: Project
        Current project.

    job_name: str
        Code generation name.
    """

    def __init__(self, project: Project, job_name: str) -> None:
        self._project = project
        self._job_name = job_name
        self._mapping = None
        self._generated_code_dir = None
        self._root_operators = []
        self._job_path = Path("")
        self._mapping_path = Path("")
        self._parse_job()

    def _parse_job(self) -> None:
        # Check and retrieve job path.
        found_job = None
        for job in self._project.directory.glob("jobs/*/.sjob"):
            with open(job) as f:
                jj = json.load(f)
                if (
                    "Properties" in jj
                    and jj.get("Kind", "") == "CodeGeneration"
                    and self._job_name == jj["Properties"].get("Name", "__not_found__")
                ):
                    self._root_operators = jj["Properties"].get("RootDeclarations", [])
                    found_job = job
                    break
        if found_job is None:
            raise ScadeOneException(
                f'Generated code: no CodeGeneration kind job named "{self._job_name}"'
            )
        self._job_path = found_job
        self._mapping_path = self._job_path.parent / "out" / "cg_map.json"

    def _load_mapping(self) -> None:
        # According to job, loads the mapping file as a dictionary.
        if self.job_path is not None:
            if not self.is_code_generated:
                raise ScadeOneException(
                    f"Generated code: code is not generated for job {self._job_name}"
                )
            try:
                with open(self._mapping_path) as f:
                    self._mapping = json.load(f)
                    self._generated_code_dir = self._mapping_path.parent / "code"
            except Exception:
                raise ScadeOneException(
                    "Generated code: cannot open mapping file"
                    + f" ({self._mapping_path.name}) for job {self._job_name}"
                )

    def _get_links(self) -> dict:
        _rtn = {}
        for m in self.mapping.get("mapping", []):
            role = m.get("role", None)
            key = (m["code_id"], role) if role else m["code_id"]
            _rtn[key] = m["model_id"]
        return _rtn

    @staticmethod
    def _capitalize_string(inp: str) -> str:
        # Capitalize the first character of string
        if "_" in inp:
            return "".join([_elm.title() for _elm in inp.split("_")])
        else:
            return inp.title()

    @staticmethod
    def _capitalize_string(inp: str) -> str:
        # Capitalize the first character of string
        if "_" in inp:
            return "".join([_elm.title() for _elm in inp.split("_")])
        else:
            return inp.title()

    @property
    def job_path(self) -> Path:
        # Returns job path.
        return self._job_path

    @property
    def mapping(self) -> dict:
        """Returns the mapping file as a dictionary."""
        if self._mapping is None:
            self._load_mapping()
        return self._mapping

    @property
    def generated_code_dir(self) -> str:
        """Returns the generated code directory path name."""
        if self._generated_code_dir is None:
            self._load_mapping()
        return self._generated_code_dir

    @property
    def root_operators(self) -> List[str]:
        """Returns the list of root operators names."""
        return self._root_operators

    @property
    def code(self) -> List[dict]:
        """Returns the 'code' data section."""
        return self.mapping["code"]

    @property
    def is_code_generated(self) -> bool:
        """Returns True if the code is generated for the job."""
        return self._mapping_path.exists()

    def decompose_code_type(self, ctype: dict) -> dict:
        """
        Returns (as a dictionary) the C type name, category (array, struct, enum, union, typedef,
        predefined_type) and sub elements of a given type element.

        The sub elements depends on the category:

        - For an array, it is a dictionary with base_type giving the type data of the element
          (same format as this method) and size giving the size of the array.
        - For a struct, it is a list of dictionaries, with name giving the field name,
          type giving the field type data (same format as this method), pointer indicating if field
          is a pointer type and size giving optional list of integer or constant names.
        - For an enum, it is a dictionary with tag_name giving internal tag name for the enum,
          and values giving the list of names corresponding to the enum values.
        - For a union, it is a list of dictionaries, with name giving the variant field name,
          type giving the variant field type data (same format as this method)
          and enum_value giving optional integer value.
        - For a typedef (imported type), it is the function names used to manipulate
          the imported type.
        - For a predefined_type, it is None.

        """
        category = ctype.get("category", "")
        content = ctype.get("content", {})
        if category == "array":
            type_elem = self.decompose_code_type(self.get_code_elem(content.get("base_type", -1)))
            type_content = {"base_type": type_elem, "size": content.get("size", 0)}
        elif category == "struct":
            type_content = []
            for e in content.get("fields", []):
                type_content.append(
                    {
                        "name": e["name"],
                        "type": self.decompose_code_type(self.get_code_elem(e["type"])),
                        "pointer": e.get("pointer", False),
                        "size": e.get("size", []),
                    }
                )
        elif category == "enum":
            elems = []
            for e in content.get("values", []):
                elems.append(e["name"])
            type_content = {"tag_name": content.get("tag_name", ""), "values": elems}
        elif category == "union":
            type_content = []
            for e in content.get("variants", []):
                type_content.append(
                    {
                        "name": e["name"],
                        "type": self.decompose_code_type(self.get_code_elem(e["type"])),
                        "enum_value": e["enum_value"],
                    }
                )
        elif category == "typedef":
            # typedef is used for imported types
            # check that the 3 functions/macros are present: init, cp, eq
            type_content = []
        elif category == "predefined_type":
            type_content = None
        else:
            type_content = content

        return {
            "name": content.get("name", ""),
            "category": category,
            "elements": type_content,
        }

    def get_code_elem(self, code_id: int) -> dict:
        """
        Returns (as a dictionary) the code data corresponding to given code identifier.

        - code_index: position of the element in the code section list
        - parent_category: category (declarations, function, struct, enum, union) of the parent
          code element
        - parent_name: name of the parent code element
        - parent_id: code identifier of the parent code element
        - category: category (declarations, function, struct, enum, union) of the code element
        - content: data content (the element itself)
        - index: position of the element in the declaration
        """

        def _parse_sub_elem(parent: CodeElem, ename: str) -> dict:
            for e in parent[1].get(ename, []):
                if e["id"] == code_id:
                    return {
                        "code_index": code_idx,
                        "parent_category": parent[0],
                        "parent_name": d[1]["name"],
                        "parent_id": d[1]["id"],
                        "category": ename,
                        "content": e,
                        "index": idx,
                    }
            return {}

        for code_idx, cont in enumerate(self.code):
            for idx, d in enumerate(cont.get("declarations", [])):
                if d[1]["id"] == code_id:
                    return {
                        "code_index": code_idx,
                        "parent_category": "declarations",
                        "parent_name": "",
                        "parent_id": -1,
                        "category": d[0],
                        "content": d[1],
                        "index": idx,
                    }
                sub = {}
                if d[0] == "function":
                    sub = _parse_sub_elem(d, "parameters")
                elif d[0] == "struct":
                    sub = _parse_sub_elem(d, "fields")
                elif d[0] == "enum":
                    sub = _parse_sub_elem(d, "values")
                elif d[0] == "union":
                    sub = _parse_sub_elem(d, "variants")
                if len(sub):
                    return sub
        return {}

    def get_field_value(self, dta: dict, field: str) -> Union[str, List[str]]:
        """
        Get value(s) of the field from a dict with its given name

        Parameters
        ----------
        dta : dict
            The input data
        field : str
            Name of field

        Returns
        -------
        Union[str, List[str]]
            Return value(s) of the field
        """
        if field in dta["content"]:
            return dta["content"][field]
        else:
            return ""

    def get_code_id(self, model_id: int, role: str = "", silent: bool = False) -> int:
        """
        Returns the code identifier associated to a given model identifier for a given *role*.

        If *silent* is set to True, the functions returns -1 in case no code is found. Otherwise,
        exception is raised.
        """
        xx = (x[0] for x in self.get_code_ids(model_id) if x[1] == role)
        try:
            return next(xx)
        except StopIteration:
            if silent:
                return -1
            else:
                role_str = f"for role {role}" if role != "" else "without role"
                raise ScadeOneException(
                    f"Generated code: no code id associated to model id #{model_id} {role_str}"
                )

    def get_code_ids(self, model_id: int) -> Generator[Tuple[int, str], None, None]:
        """
        Returns the list of code identifiers associated to a given model identifier
        (as a Generator of code, role tuples).
        """
        return (
            (x["code_id"], x.get("role", ""))
            for x in self.mapping["mapping"]
            if x["model_id"] == model_id
        )

    def get_model_elements(self, filter: str = "") -> Generator[MapElem, None, None]:
        """
        Returns the list of model elements defined in the mapping file
        (as a Generator of MapElem objects).

        If *filter* is set, returns only the elements of the given category
        (elaboration, predefined_type, array, struct, enum, variant, named_type, sensor,
        operator, mono).
        """
        return (y for y in self.mapping["model"] if filter == "" or filter == y[0])

    def get_model_id_mapping(self) -> dict:
        """
        Get model mapping

        Returns
        -------
        dict
            Return a dict with key is ID and values are a tuple of data type and parameters
        """
        _mapping = {}
        for _dcl in self.mapping["model"]:
            _cls, _params = _dcl
            _cvt = GeneratedCode._capitalize_string(_cls)
            if _cvt in ModelKeyword.__members__:
                _key = ModelKeyword[_cvt]
                if _key == ModelKeyword.Enum:
                    _mapping[_params["id"]] = [_key, _params]
                    for v in _params.get("values", []):
                        _mapping[v["id"]] = [
                            ModelKeyword.to_str(ModelKeyword.EnumValue),
                            v,
                        ]
                elif _key == ModelKeyword.Struct:
                    _mapping[_params["id"]] = [_key, _params]
                    for f in _params.get("fields", []):
                        _mapping[f["id"]] = [ModelKeyword.to_str(ModelKeyword.Field), f]
                # elif _key == ModelKeyword.Array: # TODO arrays
                elif _key == ModelKeyword.Operator:
                    _mapping[_params["id"]] = [_key, _params]
                    for io in _params.get("inputs", []):
                        _mapping[io["id"]] = [
                            ModelKeyword.to_str(ModelKeyword.Inputs),
                            io,
                        ]
                    for io in _params.get("outputs", []):
                        _mapping[io["id"]] = [
                            ModelKeyword.to_str(ModelKeyword.Outputs),
                            io,
                        ]
                else:
                    _mapping[_params["id"]] = [_key, _params]
        return _mapping

    def get_method_data(self, id_model: int, method_name: str) -> dict:
        """
        Get data of a method with its name given

        Parameters
        ----------
        id_model : int
            ID of model
        method_name : str
            Name of model

        Returns
        -------
        dict
            Return a dict of method data
        """
        _rtn = {}
        _id_method = self.get_code_id(id_model, method_name, True)

        if _id_method != -1:
            _rtn = self.get_code_elem(_id_method)
        return _rtn

    def get_operator_name(self, id_model: int) -> str:
        """
        Get an operator name

        Parameters
        ----------
        id_model : int
            ID of model

        Returns
        -------
        str
            Name of operator if it's found in the model
        """
        _rtn = ""
        if self.get_code_elem(id_model):
            _rtn = self.get_code_elem(id_model)["content"]["name"]
        return _rtn

    def get_role_type(self, id_model: int, id_code: int) -> str:
        """
        Get the role of a type with its ID given

        Parameters
        ----------
        id_model : int
            ID of model
        id_code : int
            ID of type given

        Returns
        -------
        str
            Returns a role name
        """
        _rtn = ""
        for _elm in self.mapping["mapping"]:
            if _elm["model_id"] == id_model and _elm["code_id"] == id_code:
                if "role" in _elm:
                    _rtn = _elm["role"]
                break
        return _rtn

    def get_function_ios(self, id_model: int, role: str) -> tuple[List[str], str]:
        """
        Get inputs/outputs of a function

        Parameters
        ----------
        id_model : int
            ID of model

        Returns
        -------
        tuple[List[str], List[str]]
            Return a tuple including list of inputs and list of outputs
        """
        _in = []
        _ctx = ""
        for _elm in self.get_code_elem(id_model)["content"]["parameters"]:
            _rol = self.get_role_type(id_model, _elm["type"])
            if _rol and _rol == role:
                _ctx = _elm["name"]
            else:
                _in.append(_elm["name"])
        return _in, _ctx

    def get_type(self, dat: List, id_type: int) -> TypeBase:
        """
        Get type data with its ID given

        Parameters
        ----------
        dat : List
            A list of data
        id_type : int
            Type data ID

        Returns
        -------
        TypeBase
            An instance of TypeBase
        """
        _rtn = None
        c_ek, c_atts = dat
        _model_mapping = self.get_model_id_mapping()
        _link = self._get_links()
        m_type_decl = _model_mapping.get(id_type)
        if not m_type_decl:
            m_type_decl = _model_mapping.get(_link.get(id_type))
        if c_ek == "predefined_type":
            _cvt = "PredefinedType"
        elif c_ek == "enum_value":
            _cvt = "EnumValue"
        else:
            _cvt = GeneratedCode._capitalize_string(c_ek)
        _key = ModelKeyword[_cvt] if _cvt in ModelKeyword.__members__ else _cvt
        assert not m_type_decl or m_type_decl[0] == _key
        m_atts = m_type_decl[1] if m_type_decl else {}
        if _key == ModelKeyword.PredefinedType:
            _rtn = Scalar(ModelKeyword.to_str(_key), m_atts.get("name", ""), c_atts["name"])
        elif _key == ModelKeyword.Enum:
            _rtn = Enumerate(
                enum_name=c_atts["name"],
                m_name=m_atts.get("name", "int32"),
                c_name="swan_int32",
                path=m_atts.get("path"),
            )
        elif _key == ModelKeyword.Array:
            base_type = self._get_array_base_type(c_atts["id"])
            _rtn = Array(
                m_name=base_type["m_name"],
                c_name=base_type["c_name"],
                path=c_atts["name"],
                sizes=base_type["sizes"],
            )
        elif _key == ModelKeyword.Struct:
            fields = []
            for f in c_atts.get("fields", []):
                new_dat = ["predefined_type", {"id": f["type"], "name": f["name"]}]
                fields.append(self.get_type(new_dat, f["type"]))
            _rtn = Structure(
                fields,
                ModelKeyword.to_str(_key),
                m_atts.get("name", ""),
                c_atts["name"],
            )
        else:
            # TODO: support others
            LOGGER.error(f"{ModelKeyword.to_str(_key)} not yet supported")
        return _rtn

    def _get_array_base_type(self, id_: int) -> dict[str, Union[str, list[int]]]:
        """
        Get a base type of array

        Parameters
        ----------
        id_ : int
            ID of an array

        Returns
        -------
        dict[str, Union[str, list[int]]]
            Return a dictionary with code name, model name and sizes
        """
        sizes = []
        return self._get_array_base_type_with_sizes(id_, sizes)

    def _get_array_base_type_with_sizes(
        self, id_: int, sizes: list[int]
    ) -> dict[str, Union[str, list[int]]]:
        """
        Get a base type of array with sizes

        Parameters
        ----------
        id_ : int
            ID of an array
        sizes : list[int]
            List of sizes

        Returns
        -------
        dict[str, Union[str, list[int]]]
            Return a dictionary with keys are c_name, m_name and sizes
        """
        _base_type = self.get_code_elem(id_)
        if _base_type["category"] == "array":
            sizes.insert(0, _base_type["content"]["size"])
            return self._get_array_base_type_with_sizes(_base_type["content"]["base_type"], sizes)
        model_id = self.get_model_id(id_)
        model_name = self.get_model_name(model_id)
        return {
            "c_name": _base_type["content"]["name"],
            "m_name": model_name,
            "sizes": sizes,
        }

    def get_model_id(self, code_id: int) -> Union[int, None]:
        try:
            map_elt = next(
                filter(lambda mapping: mapping["code_id"] == code_id, self.mapping["mapping"])
            )
            return map_elt["model_id"]
        except StopIteration:
            return None

    def get_model_name(self, model_id: int) -> str:
        try:
            model_elt = next(filter(lambda elt: elt[1]["id"] == model_id, self.mapping["model"]))
            return model_elt[1]["name"]
        except StopIteration:
            return ""

    def get_type_id_mapping(self) -> dict:
        """
        Get type mapping

        Returns
        -------
        dict
            Return a dict where key is ID and values are a tuple of data type and parameters
        """
        _mapping = {}
        for file in self.mapping["code"]:
            for decl in file["declarations"]:
                _cls, atts = decl
                _cvt = GeneratedCode._capitalize_string(_cls)
                if _cvt in ModelKeyword.__members__:
                    _key = ModelKeyword[_cvt]
                    if _key in {
                        ModelKeyword.PredefinedType,
                        ModelKeyword.Array,
                        ModelKeyword.Enum,
                        ModelKeyword.Struct,
                        ModelKeyword.Global,
                    }:
                        _mapping[atts["id"]] = [ModelKeyword.to_str(_key), atts]
        return _mapping

    def get_interface_file(self, id_model: int) -> dict:
        """
        Returns (as a dictionary) the interface file data corresponding to given model identifier.

        Parameters
        ----------
        id_model : int
            ID of model

        Returns
        -------
        dict
            Returns the interface file data
        """
        _rtn = {}
        for _elm in self.code:
            for _dct in _elm.get("declarations", []):
                if _dct[1]["id"] == id_model:
                    _rtn = _elm
                    break
        return _rtn

    @functools.cache
    def get_model_monomorphic_instance(self, op_path: str) -> ModelMonomorphicInstance:
        """Returns the *ModelMonomorphicInstance* object corresponding
        to given operator path name."""
        try:
            op_elem = next(i for _, i in self.get_model_elements("mono") if i["path"] == op_path)
            return ModelMonomorphicInstance(self, op_elem)
        except StopIteration:
            raise ScadeOneException(f"Generated code: no monomorphic instance named {op_path}")

    @functools.cache
    def get_model_monomorphic_instances(self) -> List[ModelMonomorphicInstance]:
        """Returns the list of *ModelMonomorphicInstance* objects for the mapping file."""
        oper_list = []
        for _, op_elem in self.get_model_elements("mono"):
            oper_list.append(ModelMonomorphicInstance(self, op_elem))
        return oper_list

    @functools.cache
    def get_model_operator(self, op_path: str) -> ModelOperator:
        """Returns the *ModelOperator* object corresponding to given operator path name."""
        try:
            op_elem = next(
                i for _, i in self.get_model_elements("operator") if i["path"] == op_path
            )
            return ModelOperator(self, op_elem)
        except StopIteration:
            raise ScadeOneException(f"Generated code: no operator named {op_path}")

    @functools.cache
    def get_model_operators(self) -> List[ModelOperator]:
        """Returns the list of *ModelOperator* objects for the mapping file."""
        return [ModelOperator(self, op_elem) for _, op_elem in self.get_model_elements("operator")]

    @functools.cache
    def get_model_sensor(self, sensor_path: str) -> ModelSensor:
        """Returns the *ModelSensor* object corresponding to given sensor path name."""
        try:
            sensor_elem = next(
                i for _, i in self.get_model_elements("sensor") if i["path"] == sensor_path
            )
            return ModelSensor(self, sensor_elem)
        except StopIteration:
            raise ScadeOneException(f"Generated code: no sensor named {sensor_path}")

    @functools.cache
    def get_model_sensors(self) -> List[ModelSensor]:
        """Returns the list of *ModelSensor* objects for the mapping file."""
        return [
            ModelSensor(self, sensor_elem) for _, sensor_elem in self.get_model_elements("sensor")
        ]

    @functools.cache
    def get_elaboration_function(self) -> Optional["CFunction"]:
        """Returns the elaboration function C function for the operator (as a CFunction object),
        if any."""
        elaboration = next(self.get_model_elements("elaboration"))
        if not elaboration:
            return None
        method_elem = self.get_code_elem(elaboration[1]["id"])
        return (
            CFunction(self, method_elem["content"], method_elem["code_index"])
            if method_elem
            else None
        )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cSpell:ignore strerror
from enum import Enum, auto
import json
from pathlib import Path
from typing import List, Optional, Union

from jsonschema import Draft7Validator
from jsonschema.exceptions import SchemaError, ValidationError

from ansys.scadeone.core import ScadeOneException
from ansys.scadeone.core.common.logger import LOGGER
from ansys.scadeone.core.common.versioning import FormatVersions

_SCHEMA = Path(__file__).parents[2] / "libs/test-results-schema.json"


class BaseTestResults:
    __test__ = False


class TestResultsParser(BaseTestResults):
    """Parse the Test results file."""

    @staticmethod
    def load(file: Union[str, Path]) -> Optional["TestResults"]:
        """Read a test results file (.JSON), create and fill a TestResults object.

        Parameters
        ----------
        file
            Test results file path.

        Returns
        -------
        TestResults
            Test results object.

        Raises
        ------
        ScadeOneException
            - Error when file is not found
            - Parse error
        """
        model = TestResultsParser._load_json_file(file)
        if not TestResultsParser._is_valid(model):
            raise ScadeOneException(f"Invalid test results file: {file}")
        if model == {}:
            return None
        return TestResultsParser._create_test_results(model)

    @staticmethod
    def _load_json_file(file: Union[str, Path]) -> dict:
        """Load the JSON file."""
        try:
            if isinstance(file, str):
                try:
                    return json.loads(file)
                except Exception as e1:
                    LOGGER.warning(f"String is not a JSON string: {e1}")
            with open(file, "r") as f:
                return json.load(f)
        except OSError as e2:
            LOGGER.error(f"Error loading test results file: {e2}")
            raise ScadeOneException(f"Error loading test results file: {e2.strerror}")

    @staticmethod
    def _is_valid(model) -> bool:
        """Validate the JSON model against the test results' schema."""
        try:
            with open(_SCHEMA, "r") as f:
                schema = json.load(f)
            Draft7Validator.check_schema(schema)
            validator = Draft7Validator(schema)
            validator.validate(model)
        except OSError as e:
            LOGGER.error(f"Error loading test results schema: {e}")
            return False
        except SchemaError as e:
            LOGGER.error(f"Error validating test results schema: {e}")
            return False
        except ValidationError as e:
            LOGGER.error(f"Error validating test results file: {e}")
            return False
        return True

    @staticmethod
    def _create_test_results(model) -> "TestResults":
        """Create the TestResults object."""
        test_results = TestResults()
        test_results._version = model.get("version")
        TestResultsParser._set_test_cases(model, test_results)
        return test_results

    @staticmethod
    def _set_test_cases(model, test_results: "TestResults") -> None:
        """Set the test cases of the TestResults object."""
        if model.get("test_cases") is None:
            return
        for test_case in model.get("test_cases"):
            tc = TestCase()
            tc._harness = test_case.get("harness")
            tc._start = test_case.get("start")
            tc._end = test_case.get("end")
            tc._status = TestResultsParser._create_test_status(test_case.get("status"))
            tc._cycles_count = test_case.get("cycles_count")
            tc._test_items = TestResultsParser._create_test_items(test_case)
            test_results._test_cases.append(tc)

    @staticmethod
    def _create_test_status(status: str) -> Optional["TestStatus"]:
        if status == "passed":
            return TestStatus.Passed
        elif status == "failed":
            return TestStatus.Failed
        elif status == "error":
            return TestStatus.Error

    @staticmethod
    def _create_test_items(test_case) -> List["TestItem"]:
        """Create the TestItem objects of the test case."""
        test_items = []
        if test_case.get("test_items") is None:
            return test_items
        for test_item in test_case.get("test_items"):
            ti = TestItem()
            ti._kind = TestResultsParser._create_test_item_kind(test_item.get("kind"))
            ti._model_path = test_item.get("model_path")
            ti._passed_count = test_item.get("passed_count")
            ti._failures = TestResultsParser._create_failures(test_item)
            test_items.append(ti)
        return test_items

    @staticmethod
    def _create_test_item_kind(kind: str) -> Optional["TestItemKind"]:
        if kind == "assert":
            return TestItemKind.Assert
        elif kind == "oracle":
            return TestItemKind.Oracle

    @staticmethod
    def _create_failures(test_item) -> List["Failure"]:
        """Create the Failure objects of the test item."""
        failures = []
        if test_item.get("failures") is None:
            return failures
        for failure in test_item.get("failures"):
            f = Failure()
            f._cycle = failure.get("cycle")
            f._actual = failure.get("actual")
            f._expected = failure.get("expected")
            f._float32_atol = failure.get("float32_atol")
            f._float32_rtol = failure.get("float32_rtol")
            f._float64_atol = failure.get("float64_atol")
            f._float64_rtol = failure.get("float64_rtol")
            f._parts_error_paths = failure.get("parts_error_paths")
            failures.append(f)
        return failures


class TestResults(BaseTestResults):
    """Test results object."""

    def __init__(self) -> None:
        self._version = None
        self._test_cases = []

    @property
    def version(self) -> str:
        """Returns the test results version."""
        return self._version

    def check_version(self):
        """Check the version of the test results.

        Raises
        ------

        ScadeOneException
            If the version is not supported
        """
        FormatVersions.check("test_results", self.version)

    @property
    def test_cases(self) -> List["TestCase"]:
        """Returns the list of test cases."""
        return self._test_cases


class TestCase(BaseTestResults):
    """Test case object."""

    def __init__(self) -> None:
        self._harness = None
        self._start = None
        self._end = None
        self._status = None
        self._cycles_count = None
        self._test_items = []

    @property
    def harness(self) -> str:
        """Returns the test harness namespace."""
        return self._harness

    @property
    def start(self) -> str:
        """Returns the test case start time."""
        return self._start

    @property
    def end(self) -> str:
        """Returns the test case end time."""
        return self._end

    @property
    def status(self) -> "TestStatus":
        """Returns the test case status."""
        return self._status

    @property
    def cycles_count(self) -> int:
        """Returns the test case cycles count."""
        return self._cycles_count

    @property
    def test_items(self) -> List["TestItem"]:
        """Returns the list of test items."""
        return self._test_items


class TestStatus(BaseTestResults, Enum):
    """Test status."""

    Passed = auto()
    """Test passed."""
    Failed = auto()
    """Test failed."""
    Error = auto()
    """Test error."""


class TestItem(BaseTestResults):
    """Test item object."""

    def __init__(self) -> None:
        self._kind = None
        self._model_path = None
        self._passed_count = None
        self._failures = []

    @property
    def kind(self) -> "TestItemKind":
        """Returns the test item kind."""
        return self._kind

    @property
    def model_path(self) -> str:
        """Returns the test item model path."""
        return self._model_path

    @property
    def passed_count(self) -> int:
        """Returns the test item passed count."""
        return self._passed_count

    @property
    def failures(self) -> List["Failure"]:
        """Returns the list of failures."""
        return self._failures


class TestItemKind(BaseTestResults, Enum):
    """Test item kind."""

    Assert = auto()
    """Assert test item."""
    Oracle = auto()
    """Oracle test item."""


class Failure:
    """Failure object."""

    def __init__(self) -> None:
        self._cycle = None
        self._actual = None
        self._expected = None
        self._float32_atol = None
        self._float32_rtol = None
        self._float64_atol = None
        self._float64_rtol = None
        self._parts_error_paths = []

    @property
    def cycle(self) -> int:
        """Returns the failure cycle."""
        return self._cycle

    @property
    def actual(self) -> str:
        """Returns the failure actual value."""
        return self._actual

    @property
    def expected(self) -> str:
        """Returns the failure expected value."""
        return self._expected

    @property
    def float32_atol(self) -> Union[int, float]:
        """Returns the absolute tolerance applied in float32 comparisons."""
        return self._float32_atol

    @property
    def float32_rtol(self) -> Union[int, float]:
        """Returns the relative tolerance applied in float32 comparisons."""
        return self._float32_rtol

    @property
    def float64_atol(self) -> Union[int, float]:
        """Returns the absolute tolerance applied in float64 comparisons."""
        return self._float64_atol

    @property
    def float64_rtol(self) -> Union[int, float]:
        """Returns the relative tolerance applied in float64 comparisons."""
        return self._float64_rtol

    @property
    def parts_error_paths(self) -> List[str]:
        """Returns the list of parts error paths."""
        return self._parts_error_paths



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains the classes for global definitions:
constants (const) and sensors.
"""

from typing import Optional, Union

import ansys.scadeone.core.swan.common as common


class ConstDecl(common.Declaration):  # numpydoc ignore=PR01
    """Constant declaration, with an id, a type, and an optional expression."""

    def __init__(
        self,
        id: common.Identifier,
        type: Optional[common.TypeExpression],
        value: Optional[common.Expression] = None,
    ) -> None:
        super().__init__(id)
        self._type_expr = type
        self._value = value

    @property
    def type(self) -> Union[common.TypeExpression, None]:
        """Type of constant."""
        return self._type_expr

    @property
    def value(self) -> Union[common.Expression, None]:
        """Constant optional value. None if undefined."""
        return self._value


class SensorDecl(common.Declaration):  # numpydoc ignore=PR01
    """Sensor declaration with an id and a type."""

    def __init__(self, id: common.Identifier, type: common.TypeExpression) -> None:
        super().__init__(id)
        self._type = type

    @property
    def type(self) -> common.TypeExpression:
        """Sensor type."""
        return self._type



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-FileCopyrightText: 2022 - 2024 ANSYS, Inc.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains the classes for operator and signature (operator
without body)
"""

from typing import Callable, List, Optional, Union, cast

from ansys.scadeone.core.svc.swan_creator.operator_creator import (
    SignatureCreator,
    OperatorCreator,
)
import ansys.scadeone.core.swan.common as common

from .diagram import Diagram
from .scopes import Scope
from .typedecl import VariableTypeExpression


class TypeConstraint(common.SwanItem):  # numpydoc ignore=PR01
    """Type constraint for operator. A constraint is:

    *where_decl* ::= **where** *typevar* {{ , *typevar* }} *numeric_kind*

    The *typevar* list can be protected and represented with string.
    """

    def __init__(
        self,
        type_vars: Union[List[VariableTypeExpression], str],
        kind: common.NumericKind,
    ) -> None:
        super().__init__()
        self._is_protected = isinstance(type_vars, str)
        self._type_vars = type_vars
        self._kind = kind

    @property
    def is_protected(self) -> bool:
        """True when types are protected."""
        return self._is_protected

    @property
    def type_vars(self) -> Union[List[VariableTypeExpression], str]:
        """Return type variable names of constraints.

        Returns
        -------
        Union[List[VariableTypeExpression], str]
            Returns the list of type names, if not protected, or
            the constraint names as a string.
        """
        return self._type_vars

    @property
    def kind(self) -> common.NumericKind:
        """Constraint numeric kind."""
        return self._kind


class OperatorSignatureBase(common.Declaration, common.ModuleItem):  # numpydoc ignore=PR01
    """Base class for Operator and Signature, gathering all interface details."""

    def __init__(
        self,
        id: common.Identifier,
        is_inlined: bool,
        is_node: bool,
        inputs: List[common.Variable],
        outputs: List[common.Variable],
        sizes: Optional[List[common.Identifier]] = None,
        constraints: Optional[List[TypeConstraint]] = None,
        specialization: Optional[common.PathIdentifier] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        common.Declaration.__init__(self, id)
        self._is_node = is_node
        self._is_inlined = is_inlined
        self._inputs = inputs
        self._outputs = outputs
        self._sizes = sizes if sizes else []
        self._constraints = constraints if constraints else []
        self._specialization = specialization
        self._pragmas = pragmas if pragmas else []
        for children in (self._inputs, self._outputs, self._sizes, self._constraints):
            self.set_owner(self, children)
        self._is_text = False

    @property
    def is_node(self) -> bool:
        """True when operator is a node."""
        return self._is_node

    @property
    def is_inlined(self) -> bool:
        """True when operator is marked for inlining."""
        return self._is_inlined

    @property
    def inputs(self) -> List[common.Variable]:
        """Return inputs as a list."""
        return self._inputs

    @property
    def outputs(self) -> List[common.Variable]:
        """Return outputs as a list."""
        return self._outputs

    @property
    def sizes(self) -> List[common.Identifier]:
        """Return sizes as a list."""
        return self._sizes

    @property
    def constraints(self) -> List[TypeConstraint]:
        """Return constraints as a list."""
        return self._constraints

    @property
    def specialization(self) -> Union[common.PathIdentifier, None]:
        """Return specialization path_id or None."""
        return self._specialization

    @property
    def pragmas(self) -> List[common.Pragma]:
        """Return pragmas as a list."""
        return self._pragmas

    @property
    def is_text(self) -> bool:
        """True when operator is given from {text%...%text} markup,
        or an interface is given from {text%...%text} markup (body)
        or {signature%...%signature} markup (interface).
        """
        return self._is_text

    @is_text.setter
    def is_text(self, text_flag: bool):
        self._is_text = text_flag


class Signature(OperatorSignatureBase, SignatureCreator):
    """Operator interface definition.

    Used in module body or interface.
    """

    def __init__(
        self,
        id: common.Identifier,
        is_inlined: bool,
        is_node: bool,
        inputs: List[common.Variable],
        outputs: List[common.Variable],
        sizes: Optional[List[common.Identifier]] = None,
        constraints: Optional[List[TypeConstraint]] = None,
        specialization: Optional[common.PathIdentifier] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        OperatorSignatureBase.__init__(
            self,
            id,
            is_inlined,
            is_node,
            inputs,
            outputs,
            sizes,
            constraints,
            specialization,
            pragmas,
        )


class Operator(OperatorSignatureBase, OperatorCreator):
    """Operator definition, with a body.

    Used in modules. The body may be empty.
    """

    def __init__(
        self,
        id: common.Identifier,
        is_inlined: bool,
        is_node: bool,
        inputs: List[common.Variable],
        outputs: List[common.Variable],
        body: Union[Scope, common.Equation, Callable] = None,
        sizes: Optional[List[common.Identifier]] = None,
        constraints: Optional[List[TypeConstraint]] = None,
        specialization: Optional[common.PathIdentifier] = None,
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        OperatorSignatureBase.__init__(
            self,
            id,
            is_inlined,
            is_node,
            inputs,
            outputs,
            sizes,
            constraints,
            specialization,
            pragmas,
        )
        self._body = body

    @property
    def body(self) -> Union[Scope, common.Equation, None]:
        """Operator body: a scope, an equation, or None."""
        if isinstance(self._body, Callable):
            body = self._body(self)
            self._body = body
            self.set_owner(self, self._body)
        return self._body

    @property
    def has_body(self) -> bool:
        """True when operator has a body."""
        return self._body is not None

    @property
    def is_equation_body(self) -> bool:
        """True when body is reduced to a single equation."""
        return isinstance(self.body, common.Equation)

    @property
    def diagrams(self) -> List[Diagram]:
        """Return a list of diagram declarations."""
        if not self.has_body or self.is_equation_body:
            return []
        return [
            cast(Diagram, diag)
            for diag in filter(lambda x: isinstance(x, Diagram), self.body.sections)
            if diag
        ]



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pathlib import Path
import re
from typing import Any, cast

import ansys.scadeone.core.svc.simdata as sd

from .sss_parser import SSSParser as Parser
from .swan_context import SwanContext
from .utils import ConverterLogger, SDType

# cSpell: ignore sdtype, sdio

SSSParser = Parser()

VarPathRE = re.compile(r"^(?:\w+::)*\w+(?:/\w+)?$")


class SDContextBase:
    def get_alias_value(self, alias: str) -> Any:
        pass

    def get_enum(self, _: str) -> Any:
        pass

    def get_constant_value(self, _: str) -> Any:
        pass

    def get_constant(self, _: str) -> Any:
        pass

    @property
    def evaluated_cmd(self) -> str:
        pass

    def failed_cmd(self, error: str):
        ConverterLogger.exception(f"Failed command: {self.evaluated_cmd}\nReason: {error}")


class Atom:
    """Stores SSS values from parsing. The values are the first element of the
     token list, then it is a value which can be a string, a number, a boolean,
    a char or a list of these types. A string between simple-quote denotes a
    character."""

    def __init__(self, tokens: list) -> None:
        # token is a list of list
        self._tokens = tokens

    @property
    def value(self):
        return self._tokens[0]

    def check_value(self, ctx: SDContextBase, sd_type: sd.Type):
        # ConverterLogger.exception if the value is not supported
        value = self._check_value(self.value, ctx, sd_type)
        self._tokens[0] = value

    def _check_array_value(self, value, ctx: SDContextBase, base_type: sd.Type, dims: list[int]):
        """Checks an array value and performs processing of "<string>".
        A value must be resolved: no alias
        """
        if len(dims) > 1:
            return [
                self._check_array_value(sub_value, ctx, base_type, dims[1:]) for sub_value in value
            ]
        if (
            isinstance(base_type, sd.PredefinedType)
            and base_type.kind == sd.PredefinedTypeKind.CHAR
            and not isinstance(value, list)
        ):
            assert isinstance(value, str) and value[0] == '"'
            chars = SSSParser.string_to_chars(value)
            if len(chars) > dims[0]:
                chars = chars[0 : dims[0]]
            else:
                chars += ["\0"] * (dims[0] - len(chars))
            return chars
        return [self._check_value(v, ctx, base_type) for v in value]

    def _check_atomic_value(self, value, ctx: SDContextBase, sd_type):
        """Checks an atomic value, i.e.
        - a numeric value,
        - a char,
        - an enum,
        - a constant expr,
        - an alias,
        - float specials (error),
        - SSS specials: ? or <empty>

        '"<string>"' value must have been processed by _check_array_value
        """
        if isinstance(value, str):
            if value == "?" or value == "<empty>":
                ctx.failed_cmd(f"Unsupported partial value ({value})")
            if value[0] == "'":
                # char
                return SSSParser.to_char(value)
            if value[0] == '"':
                assert isinstance(sd_type, sd.ArrayType)
                res = self._check_array_value(value, ctx, sd_type.base_type, sd_type.dims)
                return res
            if value in ("NaN", "qNaN", "sNaN", "+Inf", "-Inf"):
                # unsupported float values
                ctx.failed_cmd(f"Unsupported float: {value}")

            if alias := cast(Atom, ctx.get_alias_value(value)):
                alias.check_value(ctx, sd_type)
                return alias.value

            if isinstance(sd_type, sd.EnumType):
                if enum_val := ctx.get_enum(value):
                    return enum_val
                # could be a constant
                if cst_value := ctx.get_constant_value(value):
                    ctx.failed_cmd(f"Invalid enum from constant: {value} (not implemented)")
                ctx.failed_cmd(f"Invalid enum from: {value}")

            cst_value = ctx.get_constant_value(value)
            if cst_value is not None:
                return cst_value
            # Can be a char value
            if (
                isinstance(sd_type, sd.PredefinedType)
                and sd_type.kind == sd.PredefinedTypeKind.CHAR
                and len(value) == 1
            ):
                return value

            ctx.failed_cmd(f"Invalid value from: {value}")
        # other types: int, float, bool
        if isinstance(sd_type, sd.EnumType) and isinstance(value, int):
            ctx.failed_cmd(f"Integer value not supported for enum\n{ctx.evaluated_cmd}")
        return value

    def _check_value(self, value, ctx: SDContextBase, sd_type: sd.Type) -> Any:
        """Checks if the value is supported and returns it.
        Performs alias resolution.
        """
        if not isinstance(value, list):
            return self._check_atomic_value(value, ctx, sd_type)

        if isinstance(sd_type, sd.StructType):
            # value must be a list
            assert isinstance(value, list)
            field_types = [f.sd_type for f in sd_type.fields]
            res = [self._check_value(v, ctx, t) for v, t in zip(value, field_types)]
            return res

        if isinstance(sd_type, sd.EnumType):
            assert isinstance(value, list)
            return [self._check_value(v, ctx, sd_type) for v in value]

        if isinstance(sd_type, sd.ArrayType):
            res = self._check_array_value(value, ctx, sd_type.base_type, sd_type.dims)
            return res

    def __str__(self):
        return str(self.value)


class SwanVar:
    """Class to separate the module, the scope and the name of a sss variable name string"""

    def __init__(self, var: str) -> None:
        self._module = None
        self._scope = None
        name_split = var.split("::")
        if len(name_split) == 2:
            self._module = name_split[0]
        var = name_split[-1]
        name_split = var.split("/")
        if len(name_split) == 2:
            self._scope = name_split[0]
        self._name = name_split[-1]

    @property
    def module(self) -> str:
        """Gets module of swan variable"""
        return self._module

    @property
    def scope(self) -> str:
        """Gets scope of swan variable"""
        return self._scope

    @property
    def name(self) -> str:
        """Gets name of swan variable"""
        return self._name

    def __str__(self):
        return f"module {self._module}: scope {self._scope}: name {self._name}"


class SDVar:
    """
    Parent class for simdata variables state
    """

    def __init__(self, name: str, element: sd.Element) -> None:
        self._name = name
        self._next_value = None  # values to be applied next cycles apply
        self._active_cycles = 0  # number of times to be played next
        self._element = element  # simdata file element
        self._sustain = (
            0  # number of times to be played again before back to none, zero = not limit
        )

    @property
    def name(self) -> str:
        """Gets variable name"""
        return self._name

    @property
    def element(self) -> sd.Element:
        """Gets associated simdata element"""
        return self._element

    @property
    def type(self) -> sd.Type:
        """Gets associated simdata element type"""
        return self.element.sd_type if self.element else None

    @property
    def active_cycles(self) -> int:
        """Gets number of cycles left to assign in the current state"""
        return self._active_cycles

    @property
    def sustain(self) -> int:
        """Gets cycles to play in case of checks"""
        return self._sustain

    @sustain.setter
    def sustain(self, sustain: int) -> None:
        """Sets cycles to play for a check"""
        self._sustain = sustain

    @property
    def value(self):
        """Gets the next value to be assigned to variable"""
        return self._next_value

    @value.setter
    def value(self, value=None) -> None:
        """Updates the value of sd variable after appending previous values
        to file element if any.
        If no value given, will only append previous values.

        Parameters
        ----------
        value : Any, optional
            new value to affect
        """
        pass

    def cycle_start(self, past_cycles: int) -> None:
        """Defines when this element is initialized

        Parameters
        ----------
        past_cycles : int
            number of cycles spent before element initialization
        """
        pass

    def add_cycles(self, n: int) -> None:
        """Increases number of cycles to append to sd variable.

        Parameters
        ----------
        n : int
            number of new cycles
        """
        pass

    def terminate(self) -> None:
        """Final operation for variable.
        Sets remaining values to file if any"""
        pass


class SDVarSimple(SDVar):
    """Class for simple predefined type SD variables"""

    def __init__(self, name: str, element: sd.Element) -> None:
        super().__init__(name, element)

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, value=None) -> None:
        if self._active_cycles > 0:
            if self._next_value is None:
                self._element.append_nones(self._active_cycles)
            else:
                self._element.append_values([self._next_value], self._active_cycles)
            self._active_cycles = 0

        self._next_value = None if value is None else value.value

    def cycle_start(self, past_cycles: int) -> None:
        self._element.append_nones(past_cycles)

    def add_cycles(self, n: int) -> None:
        total_cycles = self._active_cycles + n
        if self._sustain and self._sustain <= total_cycles:
            self._active_cycles = self._sustain
            self.value = None

            self._active_cycles = total_cycles - self._sustain
            self._sustain = 0
        else:
            self._active_cycles = total_cycles

    def terminate(self) -> None:
        self.value = None


class SDVarStruct(SDVar):
    """Class for structured variables.
    Structured values are only assigned at the end of parsing,
    due to how structured simdata types do not support more than one sequence"""

    def __init__(self, name: str, element: sd.Element) -> None:
        super().__init__(name, element)
        self._sequence_value = []

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, value=None) -> None:
        if value is None:
            self._next_value = None
            return
        else:
            eval_value = value.value
            self._next_value = eval_value

    def cycle_start(self, past_cycles: int) -> None:
        self._sequence_value = [None] * past_cycles

    def add_cycles(self, n: int) -> None:
        self._sequence_value += [self._next_value] * n

    def terminate(self) -> None:
        if None in self._sequence_value:
            if any(self._sequence_value):
                ConverterLogger.warning(
                    "Structured types do not support combination of values and nones. "
                    "Structured values must be defined for every cycle. "
                    f"Variable {self._name} will not be initialized."
                )
            self._element.append_nones(len(self._sequence_value))
        else:
            try:
                self._element.append_values(self._sequence_value)
            except Exception as e:
                raise e


class SDVarNone(SDVar):
    """Class for unsupported SD variables.
    That is complex structure type variables or which type was not found"""

    pass


class SDContext(SDContextBase):
    """
    Context for simdata writing
    Contains current state of it: cycle number, list of SDvariables with their respective state,
    sd file and associated swan context in order to find variable types
    """

    def __init__(self, swan_ctx: SwanContext) -> None:
        self._vars: dict[str, SDVar] = {}
        self._vars_aliases = {}
        self._values_aliases: dict[str, Atom] = {}
        self._sdf = None
        self._sdio_type = None
        self._current_cycle = 0
        self._swan_ctx = swan_ctx
        self._cmds = []
        self._evaluated_cmd = None

    @property
    def evaluated_cmd(self) -> str:
        return self._evaluated_cmd

    def parse_file(self, file: Path) -> None:
        """Parse content of a file and append commands"""
        ConverterLogger.info(f"Parsing file:\n{file}")
        self._cmds.extend(SSSParser.parse_file(file))

    def reset_vars(self) -> None:
        """Resets all the simdata variables"""
        self._vars = {}
        self._vars_aliases = {}

    @property
    def sdf(self) -> sd.FileBase:
        """Gets current associated simdata file"""
        return self._sdf

    @sdf.setter
    def sdf(self, value: sd.FileBase) -> None:
        """Associates a simdata file"""
        self._sdf = value

    @property
    def sdio_type(self) -> SDType:
        """Gets conversion mode : input or output (checks)"""
        return self._sdio_type

    @sdio_type.setter
    def sdio_type(self, type: SDType) -> None:
        """Sets conversion mode : input or output"""
        self._sdio_type = type

    def init_var(self, sig_name: str, type: sd.Type, is_sensor: bool = False):
        swan_var = SwanVar(sig_name)
        short_name = swan_var.name
        element = self._sdf.add_element(
            sig_name.replace("::", "_") if is_sensor else short_name, type
        )

        if type is None:
            self._vars[sig_name] = SDVarNone(sig_name, element)
            ConverterLogger.error(f"Unsupported type: variable {sig_name} will not be converted!")
            return
        elif isinstance(type, sd.PredefinedType) or isinstance(type, sd.EnumType):
            self._vars[sig_name] = SDVarSimple(sig_name, element)
        else:
            self._vars[sig_name] = SDVarStruct(sig_name, element)

        if self._current_cycle > 0:
            self._vars[sig_name].cycle_start(self._current_cycle)

    def init_all_vars(self) -> None:
        """Inits all inputs or outputs by creating SD elements
        with their respective type"""

        if self._sdio_type == SDType.INPUT:
            items = self._swan_ctx.inputs.items()
        elif self._sdio_type == SDType.OUTPUT:
            items = self._swan_ctx.outputs.items()
        else:
            items = []
        for sig_name, type in items:
            self.init_var(sig_name, type)

    def get_var(self, sss_name: str) -> SDVar:
        """Gets an SD element from its name"""

        # check if name is an alias
        sss_name = self._vars_aliases.get(sss_name, sss_name)
        if sss_name.find("/") == -1:
            # Sensor
            resolved_name = self.get_sensor(sss_name)
        else:
            # Operator I/O, probes
            if sss_name.find("::") == -1:
                resolved_name = self._swan_ctx.operator_module + "::" + sss_name
            else:
                (path, name) = sss_name.split("/")
                resolved_path, _ = self._swan_ctx.get_renamed("operator", path)
                resolved_name = resolved_path + "/" + name

        if resolved_name not in self._vars:
            ConverterLogger.exception(f"Variable {resolved_name} not found")

        return self._vars[resolved_name]

    def get_sensor(self, name: str) -> str:
        """Gets the sensor name from the swan context and returns its full name"""
        (resolved_name, _) = self._swan_ctx.get_renamed("sensor", name)
        if resolved_name not in self._vars:
            sensor_type = self._swan_ctx.get_sensor_sd_type(resolved_name)
            if sensor_type is not None:
                self.init_var(resolved_name, sensor_type, True)
            else:
                ConverterLogger.error(f"A Swan variable could not be found: {name}")
                self._vars[resolved_name] = SDVarNone(resolved_name, None)
        return resolved_name

    def is_supported_name(self, name: str) -> bool:
        """Test if the variable name passes some checks"""

        if not name:
            ConverterLogger.error("Variable name cannot be empty")
            return False
        m = VarPathRE.match(name)
        if m is None:
            ConverterLogger.error(f"Variable name not supported: {name}")
            return False
        return True

    def get_alias_value(self, alias: str) -> Any:
        """Gets the value of an alias"""
        if alias in self._values_aliases:
            (value, complex) = self._values_aliases[alias]
            if value:
                return value
            ConverterLogger.exception(f"Unsupported alias value: {alias} {' '.join(complex)}")
        return None

    def get_enum(self, value: str) -> Any:
        """Finds the enum value from the swan context"""
        return self._swan_ctx.get_enum(value)

    def get_constant_value(self, value: str) -> Any:
        """Finds the constant value from the swan context"""
        return self._swan_ctx.get_constant_value(value, no_const_ok=True)

    def get_constant(self, value: str) -> Any:
        """Finds the constant value from the swan context"""
        return self._swan_ctx.get_constant(value)

    def eval_cmds(self) -> None:
        self.init_all_vars()
        for cmd in self._cmds:
            ConverterLogger.debug(f"Eval: {str(cmd)}")
            self._evaluated_cmd = str(cmd)
            try:
                cmd.eval(self)
            except Exception:
                pass
        self.finish()
        self.reset_vars()

    def eval_alias(self, name: str, variable: str) -> None:
        """Creates alias for a variable

        Parameters
        ----------
        name : str
            alias
        variable : str
            full name variable
        """

        if self.is_supported_name(variable):
            self._vars_aliases[name] = variable

    def eval_alias_value(self, alias: str, value: Atom, complex: str) -> None:
        """Creates alias for a value

        Parameters
        ----------
        alias : str
            alias
        value : Any
            value
        """
        self._values_aliases[alias] = (value, complex)

    def eval_set(self, name: str, value: Atom) -> None:
        """Sets a variable to a value

        Parameters
        ----------
        name : str
            variable name
        value : Any
            value with the variable type
        """
        if self.is_supported_name(name) and self._sdio_type == SDType.INPUT:
            sd_var = self.get_var(name)
            if sd_var.type is None:
                return
            value.check_value(self, sd_var.type)
            try:
                sd_var.value = value
            except Exception as e:
                self.failed_cmd(str(e))

    def eval_cycle(self, nb: int) -> None:
        """Increment cycles for all sd variables

        Parameters
        ----------
        nb : int
            number of cycles
        """

        for sig in self._vars.values():
            sig.add_cycles(nb)
        self._current_cycle += nb

    # Oracles
    def eval_check(self, name: str, value: Atom, sustain: int, real: Any) -> None:
        """Sets an oracle check

        Parameters
        ----------
        name : str
            variable name
        value : Any
            value with the check variable type
        sustain : int
            number of cycles persistence
        """

        if self.is_supported_name(name) and self._sdio_type == SDType.OUTPUT:
            if real:
                ConverterLogger.warning(f"Unsupported SSM::check real option {self.evaluated_cmd}")
            sd_var = self.get_var(name)
            if sd_var.type is None:
                return
            value.check_value(self, sd_var.type)
            try:
                sd_var.value = value
                sd_var.sustain = sustain
            except Exception as e:
                self.failed_cmd(str(e))

    def eval_uncheck(self, name: str) -> None:
        """Removes an oracle check

        Parameters
        ----------
        name : str
            check variable name
        """

        if self._sdio_type == SDType.OUTPUT:
            sd_var = self.get_var(name)
            sd_var.value = None

    def finish(self) -> None:
        """
        Final operations to apply to sd file creation
        Must be called at the end of the sss files parsing
        Fills any variable so that the number of values is equal for all of them.
        Fills oracles with nones.
        Ignore variables that never got any value.
        """

        for sd_var in self._vars.values():
            sd_var.terminate()

        self._current_cycle = 0

    def eval_unsupported(self, line: str) -> None:
        """Logs a SSM command that will not be treated

        Parameters
        ----------
        line : str
            The command line
        """
        if self._sdio_type == SDType.INPUT:
            # Arbitrary choice to avoid logging twice the same lines
            ConverterLogger.warning(f"Unsupported command: {line}")


class SSSCmd:
    """
    Parent class for SSM commands
    """

    def eval(self, context: SDContext) -> None:
        pass


class SSSCycle(SSSCmd):
    """SSM::cycle [[ <integer> ]]
    Apply cycles to registered values"""

    def __init__(self, cycle) -> None:
        super().__init__()
        self._cycle = int(cycle)

    def eval(self, context: SDContext) -> None:
        context.eval_cycle(self._cycle)

    def __str__(self) -> str:
        name = self.__class__.__name__
        return f"{name} {self._cycle}"


class SSSSet(SSSCmd):
    """SSM::set <var> <val>
    Setting values"""

    def __init__(self, signal: str, value: Atom) -> None:
        super().__init__()
        self._value = value
        self._signal = signal

    def eval(self, context: SDContext) -> None:
        context.eval_set(self._signal, self._value)

    def __str__(self) -> str:
        name = self.__class__.__name__
        return f"{name} {self._signal} {self._value}"


class SSSCheck(SSSCmd):
    """SSM::check <var> <expected val> {{ <check var arg> }
    Not supported for images check"""

    def __init__(self, signal, value: Atom, sustain, real) -> None:
        super().__init__()
        self._value = value
        self._signal = signal
        self._sustain = sustain
        self._real = real

    def eval(self, context: SDContext) -> None:
        context.eval_check(self._signal, self._value, self._sustain, self._real)

    def __str__(self) -> str:
        name = self.__class__.__name__
        real = f"real={self._real}" if self._real else ""
        if self._sustain == 0:
            sustain = "sustain=forever"
        elif self._sustain == 1:
            sustain = ""
        else:
            sustain = f"sustain={self._sustain}"
        return f"{name} {self._signal} {self._value} {sustain} {real}"


class SSSTolerance(SSSCmd):
    """SSM::set_tolerance {{ <tolerance arg> }}
    Not supported in sd files"""

    def __init__(self, line) -> None:
        super().__init__()
        self._line = line

    def eval(self, context: SDContext) -> None:
        context.eval_unsupported(f"SSM::set_tolerance{self._line}")

    def __str__(self) -> str:
        name = self.__class__.__name__
        return f"{name} {self._line}"


class SSSUncheck(SSSCmd):
    """SSM::uncheck <var>
    Not supported in sd files"""

    def __init__(self, line) -> None:
        super().__init__()
        self._line = line

    def eval(self, ctx: SDContext) -> None:
        ctx.eval_unsupported(f"SSM::uncheck {self._line}")

    def __str__(self) -> str:
        name = self.__class__.__name__
        return f"{name} {self._line}"


class SSSAlias(SSSCmd):
    """SSM::alias <alias> <var>"""

    def __init__(self, signal, value) -> None:
        super().__init__()
        self._value = value
        self._signal = signal

    def eval(self, context: SDContext) -> None:
        context.eval_alias(self._signal, self._value)

    def __str__(self) -> str:
        name = self.__class__.__name__
        return f"{name} {self._signal} {self._value}"


class SSSAliasValue(SSSCmd):
    """SSM::alias_value <alias> <var>"""

    def __init__(self, alias: str, value: Atom, complex: str) -> None:
        super().__init__()
        self._alias = alias
        self._value = value
        self._complex = complex

    def eval(self, context: SDContext) -> None:
        context.eval_alias_value(self._alias, self._value, self._complex)

    def __str__(self) -> str:
        name = self.__class__.__name__
        return f"{name} {self._alias} {self._value}"


class SSSUnsupported(SSSCmd):
    """Unsupported SSM actions"""

    def __init__(self, line) -> None:
        super().__init__()
        self._line = line

    def eval(self, context: SDContext) -> None:
        context.eval_unsupported(self._line)

    def __str__(self) -> str:
        name = self.__class__.__name__
        return f"{name} {self._line}"


SSSParser.initialize(
    atom=Atom,
    sss_set=SSSSet,
    sss_check=SSSCheck,
    sss_alias=SSSAlias,
    sss_alias_value=SSSAliasValue,
    sss_cycle=SSSCycle,
    sss_tolerance=SSSTolerance,
    unsupported=SSSUnsupported,
)
